"""pKa workflow."""

from typing import Annotated, Self

from pydantic import AfterValidator, model_validator

from ..base import Base, LowercaseStrEnum, round_float, round_optional_float
from ..mode import Mode
from ..molecule import Molecule
from ..solvent import Solvent
from .workflow import DBCalculation, MoleculeWorkflow, SMILESWorkflow

CHEMPROP_NEVOLIANIS2025_ALLOWED_SOLVENTS = {
    Solvent.WATER,
    Solvent.DIMETHYLSULFOXIDE,
    Solvent.DIMETHYLFORMAMIDE,
    Solvent.ACETONITRILE,
    Solvent.METHANOL,
    Solvent.ETHANOL,
    Solvent.ETHYLENE_GLYCOL,
    Solvent.N_METHYLPYRROLIDONE,
}


class MicroscopicpKaMethod(LowercaseStrEnum):
    AIMNET2_WAGEN2024 = "aimnet2_wagen2024"
    CHEMPROP_NEVOLIANIS2025 = "chemprop_nevolianis2025"
    STARLING = "starling"


class pKaMicrostate(Base):
    """
    A microstate for pKa calculations.

    :param atom_index: index of the atom
    :param smiles: SMILES of the microstate
    :param structures: DBCalculation for the microstate
    :param deltaG: relative free energy (where applicable)
    :param pka: pKa value associated with this microstate
    :uncertainty: uncertainty in the pKa prediction
    """

    atom_index: int
    smiles: str | None = None
    structures: list[DBCalculation] = []
    deltaG: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    pka: Annotated[float, AfterValidator(round_float(3))]
    uncertainty: Annotated[float | None, AfterValidator(round_optional_float(3))] = None


class pKaWorkflow(SMILESWorkflow, MoleculeWorkflow):
    """
    Workflow for calculating pKa.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow
    :param initial_smiles: SMILES of the molecule of interest

    New:
    :param microscopic_pka_method: method used for pka prediciton
    :param solvent: solvent for the pka prediction
    :param pka_range: range of pKa values to consider
    :param deprotonate_elements: elements to deprotonate
    :param deprotonate_atoms: atoms to deprotonate
    :param protonate_elements: elements to protonate
    :param protonate_atoms: atom indices to protonate
    :param reasonableness_buffer: buffer for pKa reasonableness

    Results:
    :param structures: DBCalculations for the microstates
    :param conjugate_acids: conjugate acid microstates
    :param conjugate_bases: conjugate base microstates
    :param strongest_acid: pKa of the strongest acid
    :param strongest_base: pKa of the strongest base
    """

    initial_smiles: str = ""
    initial_molecule: Molecule | None = None  # type: ignore [assignment]

    mode: Mode = Mode.CAREFUL

    microscopic_pka_method: MicroscopicpKaMethod = MicroscopicpKaMethod.AIMNET2_WAGEN2024
    solvent: Solvent = Solvent.WATER
    pka_range: tuple[float, float] = (2, 12)
    deprotonate_elements: list[int] = [7, 8, 16]
    deprotonate_atoms: list[int] = []
    protonate_elements: list[int] = [7]
    protonate_atoms: list[int] = []

    reasonableness_buffer: float = 5

    structures: list[DBCalculation] = []
    conjugate_acids: list[pKaMicrostate] = []
    conjugate_bases: list[pKaMicrostate] = []
    strongest_acid: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    strongest_base: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    @model_validator(mode="after")
    def check_method_settings(self) -> Self:
        """Check that models with limited domain of applicability are predicting within correct domain."""
        match self.microscopic_pka_method:
            case MicroscopicpKaMethod.AIMNET2_WAGEN2024:
                if self.solvent is not Solvent.WATER:
                    raise ValueError(f"{self.microscopic_pka_method} only supports water")
            case MicroscopicpKaMethod.CHEMPROP_NEVOLIANIS2025:
                if self.solvent not in CHEMPROP_NEVOLIANIS2025_ALLOWED_SOLVENTS:
                    raise ValueError(f"Solvent `{self.solvent}` is not supported by method `{self.microscopic_pka_method}`.")
                if len(self.protonate_atoms) or len(self.deprotonate_atoms):
                    raise ValueError(f"Method `{self.microscopic_pka_method}` does not support selecting atoms by number.")
            case MicroscopicpKaMethod.STARLING:
                if self.solvent is not Solvent.WATER:
                    raise ValueError(f"{self.microscopic_pka_method} only supports water")
                if len(self.protonate_atoms) or len(self.deprotonate_atoms):
                    raise ValueError(f"Method `{self.microscopic_pka_method}` does not support selecting atoms by number.")
        return self

    @model_validator(mode="after")
    def validate_mol_input(self) -> Self:
        """Ensure that only one of initial_molecule or initial_smiles is set."""

        if not (bool(self.initial_smiles) ^ bool(self.initial_molecule)):
            raise ValueError("Can only set one of initial_molecule and initial_smiles")

        return self



"""Protein-related data models."""

from .base import Base


class ProteinSequence(Base):
    """
    Protein sequence data.

    :param sequence: amino-acid sequence string
    :param cyclic: whether this sequence forms a cyclic peptide
    """

    sequence: str
    cyclic: bool = False
    id: str | None = None



from typing import Annotated, Self

from pydantic import AfterValidator, model_validator

from ..base import LowercaseStrEnum, round_float
from ..molecule import Molecule
from .workflow import MoleculeWorkflow, SMILESWorkflow


class MembranePermeabilityMethod(LowercaseStrEnum):
    CHEMPROP_OHLSSON2025 = "chemprop_ohlsson2025"
    PYPERMM = "pypermm"


class MembranePermeabilityWorkflow(SMILESWorkflow, MoleculeWorkflow):
    """
    Membrane permeability prediction workflow.

    Inherited:
    :param initial_smiles: SMILES string of the molecule
    :param initial_molecule: SMILES string of the molecule

    Inputs:
    :param membrane_permeability_method: model used to predict membrane permeability

    Results:
    :param caco_2_P_app: base-10 logarithm of the Caco-2 apparent permeability P_app, in cm/s
    :param caco_2_logP: Caco-2 intrinsic permeability coefficient logP
    :param blm_logP: bilayer-lipid-membrane intrinsic permeability coefficient logP
    :param pampa_logP: PAMPA intrinsic permeability coefficient logP
    :param plasma_logP: plasma-membrane intrinsic permeability coefficient logP
    :param bbb_logP: blood–brain-barrier intrinsic permeability coefficient logP
    :param energy_profile: how the energy of the compound is predicted to change as it passes through a membrane.
        the first value is the membrane profile, in Å, the second is the energy, in kcal/mol
    """

    initial_smiles: str = ""
    initial_molecule: Molecule | None = None  # type: ignore [assignment]

    membrane_permeability_method: MembranePermeabilityMethod = MembranePermeabilityMethod.CHEMPROP_OHLSSON2025

    caco_2_P_app: Annotated[float, AfterValidator(round_float(3))] | None = None

    caco_2_logP: Annotated[float, AfterValidator(round_float(3))] | None = None
    blm_logP: Annotated[float, AfterValidator(round_float(3))] | None = None
    pampa_logP: Annotated[float, AfterValidator(round_float(3))] | None = None
    plasma_logP: Annotated[float, AfterValidator(round_float(3))] | None = None
    bbb_logP: Annotated[float, AfterValidator(round_float(3))] | None = None

    energy_profile: list[
        tuple[
            Annotated[float, AfterValidator(round_float(3))],
            Annotated[float, AfterValidator(round_float(3))],
        ]
    ] = []

    @model_validator(mode="after")
    def validate_mol_input(self) -> Self:
        """Ensure that only one of initial_molecule or initial_smiles is set."""

        if not (bool(self.initial_smiles) ^ bool(self.initial_molecule)):
            raise ValueError("Can only set one of initial_molecule and initial_smiles")

        if self.initial_molecule:
            assert self.membrane_permeability_method == MembranePermeabilityMethod.PYPERMM
        elif self.initial_smiles:
            assert self.membrane_permeability_method == MembranePermeabilityMethod.CHEMPROP_OHLSSON2025

        return self



from typing import Self

from pydantic import model_validator

from .base import Base, LowercaseStrEnum, UniqueList
from .message import Message
from .molecule import Molecule
from .settings import Settings
from .status import Status
from .task import Task
from .types import UUID


class StJamesVersion(LowercaseStrEnum):
    """
    Flag for which version of stjames this calculation was saved in.
    This will become meaningful later, when we have (probably) multiple versions.
    """

    V0 = "09072023"


class Calculation(Base):
    """
    Computational chemistry calculation.

    :param molecules: molecules generated by the calculation
    :param tasks: tasks to perform (energy, optimize, etc.)
    :param settings: computational settings
    :param status: execution status
    :param name: name for calculation
    :param elapsed: wall time (s)
    :param logfile: calculation log output
    :param messages: user-facing messages
    :param engine: deprecated, use settings.engine
    :param uuids: UUIDs of calculations in database
    :param json_format: stjames version identifier
    """

    molecules: list[Molecule]

    tasks: UniqueList[Task] = []
    settings: Settings = Settings()

    status: Status = Status.QUEUED

    name: str | None = None
    elapsed: float | None = None
    logfile: str | None = None
    messages: list[Message] = []

    # DEPRECATED - moving into settings
    engine: str | None = "peregrine"

    uuids: list[UUID | None] | None = None

    # not to be changed by end users, diff. versions will have diff. defaults
    json_format: str = StJamesVersion.V0

    @model_validator(mode="after")
    def populate_tasks(self) -> Self:
        """Set the tasks from the settings, so that we don't have to migrate old entries."""
        if len(self.tasks) == 0:
            self.tasks = self.settings.tasks
        return self



from pydantic import ValidationError
from pytest import raises

from stjames import Constraint
from stjames.conformers import ConformerGenMixin, ConformerSearchMixin, ETKDGSettings, iMTDSettings, iMTDSpeeds
from stjames.method import Method


def test_etkdg() -> None:
    settings = ETKDGSettings()

    assert settings.num_initial_confs == 300

    with raises(ValidationError, match="ETKDG does not support NCI"):
        ETKDGSettings(nci=True)

    with raises(ValidationError, match="ETKDG does not support constraints"):
        ETKDGSettings(constraints=[Constraint(constraint_type="bond", atoms=[1, 2])])


def test_imtdgc() -> None:
    settings = iMTDSettings()

    assert settings.speed == iMTDSpeeds.QUICK
    assert not settings.reopt
    assert settings.mtd_method == Method.GFN_FF


def test_conformer_gen_mixin() -> None:
    settings = ConformerGenMixin(conf_gen_settings=ETKDGSettings(num_initial_confs=150))
    assert settings.conf_gen_settings == ETKDGSettings(num_initial_confs=150)


def test_conformer_search_mixin() -> None:
    settings = ConformerSearchMixin(frequencies=True, solvent="water", conf_gen_settings=ETKDGSettings())

    assert settings.solvent == "water"
    assert settings.frequencies



"""Docking workflow."""

from typing import Annotated, Any, Literal, Self, TypeAlias

from pydantic import AfterValidator, ConfigDict, PositiveInt, field_validator, model_validator

from ..base import Base, round_float
from ..conformers import ConformerGenSettingsUnion, ETKDGSettings
from ..pdb import PDB
from ..types import UUID, ProteinUUID, Vector3D
from .workflow import MoleculeWorkflow, ProteinStructureWorkflow

CalculationUUID: TypeAlias = UUID


class Score(Base):
    """
    Pose with its score.

    :param pose: conformation of the ligand when docked (calculation UUID)
    :param complex_pdb: UUID of protein–ligand complex (protein UUID)
    :param score: score of the pose, (kcal/mol)
    :param posebusters_valid: whether or not the ligand pose passes the PoseBusters tests
    :param strain: strain (kcal/mol)
    :param rmsd: RMSD from the reference, if there's a reference molecule to dock against (Å)
    """

    pose: CalculationUUID | None
    complex_pdb: ProteinUUID | None
    score: Annotated[float, AfterValidator(round_float(3))]
    posebusters_valid: bool
    strain: float | None
    rmsd: Annotated[float, AfterValidator(round_float(3))] | None = None


class DockingSettings(Base):
    """
    Base class for controlling how docked poses are generated.

    :param max_poses: maximum number of poses generated per input molecule
    """

    max_poses: int = 4


class VinaSettings(DockingSettings):
    """
    Controls how AutoDock Vina is run.

    :param executable: which Vina implementation is run.
    :param scoring_function: which scoring function is employed.
    :param exhaustiveness: how many times Vina attempts to find a pose.
        8 is typical, 32 is considered relatively careful.
    """

    executable: Literal["qvina2", "qvina-w", "vina"] = "vina"
    scoring_function: Literal["vinardo", "vina"] = "vinardo"
    exhaustiveness: int = 8

    @model_validator(mode="after")
    def check_executable_scoring_function(self) -> Self:
        """Check if the combination of exectuable and scoring function is supported."""
        if (self.executable in {"qvina2", "qvina-w"}) and (self.scoring_function == "vinardo"):
            raise ValueError("QVina does not implement the Vinardo scoring function!")
        return self


class DockingWorkflow(MoleculeWorkflow, ProteinStructureWorkflow):
    """
    Docking workflow.

    Note that the protein can be supplied either by UUID or raw PDB object.
    We anticipate that the former will dominate deployed usage, but the latter is handy for isolated testing.
    If, for whatever reason, the workflow is initialized with both a `target_uuid` and a `target`, the UUID will be ignored.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)
    :param protein: Protein target, as PDB or UUID

    New:
    :param target: PDB of the protein; DEPRECATED.
    :param target_uuid: UUID of the protein; DEPRECATED.
    :param pocket: center (x, y, z) and size (x, y, z) of the pocket
    :param do_csearch: whether to csearch starting structures
    :param conformer_gen_settings: settings for initial conformer search.
    :param do_optimization: whether to optimize starting structures
    :param do_pose_refinement: whether to optimize non-rotatable bonds in output poses

    Results:
    :param conformers: UUIDs of optimized conformers
    :param scores: docked poses sorted by score
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    target: PDB | None = None
    target_uuid: UUID | None = None
    pocket: tuple[Vector3D, Vector3D]

    docking_settings: VinaSettings = VinaSettings()

    do_csearch: bool = True
    conformer_gen_settings: ConformerGenSettingsUnion = ETKDGSettings(
        num_initial_confs=200,
        num_confs_considered=50,
        max_confs=20,
        max_mmff_energy=20,
    )
    do_optimization: bool = True
    do_pose_refinement: bool = True

    conformers: list[CalculationUUID] = []
    scores: list[Score] = []

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the Docking workflow."""
        if self.target is not None:
            desc = self.target.description
            target = desc.code or desc.title
        else:
            target = ""

        ligand = "".join(atom.atomic_symbol for atom in self.initial_molecule.atoms)
        return f"<{type(self).__name__} {target} {ligand}>"

    @model_validator(mode="before")
    def harmonize_target_and_protein(cls, data: Any) -> Any:  # noqa: N805
        """
        Syncs data between "target"/"target_uuid" and "protein" field.
        """
        protein = data.get("protein", False)
        target = data.get("target", False)
        target_uuid = data.get("target_uuid", False)

        if not protein:
            if target:
                data["protein"] = target
            elif target_uuid:
                data["protein"] = target_uuid
        elif not target and not target_uuid:
            if isinstance(data["protein"], PDB):
                data["target"] = protein
            elif isinstance(data["protein"], UUID):
                data["target_uuid"] = protein

        return data

    @model_validator(mode="after")
    def check_protein(self) -> Self:
        """Check if protein is provided."""
        if not self.target and not self.target_uuid:
            raise ValueError("Must provide either target or target_uuid")
        return self

    @field_validator("pocket", mode="after")
    def validate_pocket(cls, pocket: tuple[Vector3D, Vector3D]) -> tuple[Vector3D, Vector3D]:
        _center, size = pocket
        if any(q <= 0 for q in size):
            raise ValueError(f"Pocket size must be positive, got: {size}")
        return pocket


class AnalogueDockingWorkflow(MoleculeWorkflow, ProteinStructureWorkflow):
    """
    Workflow for docking analogues:
    (1) Conformers are generated in analogous poses to the initial molecule.
    (2) They're then optimized locally using the docking scoring function.
    (3) PoseBusters is used to check the validity of the output poses.

    Inherited:
    :param initial_molecule: molecule of interest, to which subsequent molecules will be aligned
    :param mode: Mode for workflow (currently unused)
    :param protein: PDB or UUID

    New:
    :param analogues: SMILES for analogues of `initial_molecule`
    :param num_conformers_per_analogue: the number of conformers to generate per analogue
    :param require_posebusters: filter conformers based on PoseBusters validity
    :param run_local_optimization: whether to run a local optimization in the docking pocket or just to score
    :param docking_settings: how docking should be run

    Results:
    :param analogue_scores: docked poses for each analogue of form {smiles: list[poses]}
    """

    analogues: list[str]
    num_conformers_per_analogue: PositiveInt = 1000
    require_posebusters: bool = True
    run_local_optimization: bool = True
    docking_settings: VinaSettings = VinaSettings()

    analogue_scores: dict[str, list[Score]] = {}



from .base import Base, LowercaseStrEnum


class ComputeType(LowercaseStrEnum):
    """Hardware compute type."""

    CPU = "cpu"
    GPU = "gpu"
    AUTO = "auto"


class ComputeSettings(Base):
    """
    Hardware compute settings.

    :param requested_compute_type: requested hardware type
    :param compute_type_used: actual hardware type used (set after execution)
    """

    requested_compute_type: ComputeType = ComputeType.CPU
    compute_type_used: ComputeType | None = None



from typing import Literal

from .base import LowercaseStrEnum
from .engine import Engine


class Method(LowercaseStrEnum):
    """Computational chemistry method."""

    HARTREE_FOCK = "hf"
    HF3C = "hf_3c"

    # DFT methods
    ## Pure functionals
    PBE = "pbe"
    BP86 = "bp86"
    B973C = "b97_3c"
    B97D3BJ = "b97_d3bj"
    R2SCAN = "r2scan"
    R2SCAN3C = "r2scan_3c"
    TPSS = "tpss"
    M06L = "m06l"

    ## Hybrid functionals
    PBE0 = "pbe0"
    B3LYP = "b3lyp"
    TPSSH = "tpssh"
    M06 = "m06"
    M062X = "m062x"

    ## Range-separated hybrid functionals
    CAMB3LYP = "camb3lyp"
    WB97XD3 = "wb97x_d3"
    WB97XV = "wb97x_v"
    WB97X3C = "wb97x_3c"
    WB97MV = "wb97m_v"
    WB97MD3BJ = "wb97m_d3bj"

    ## Double hybrid functionals
    DSDBLYPD3BJ = "dsd_blyp_d3bj"

    ## ML functionals
    SKALA = "skala"

    # NNPs
    AIMNET2_WB97MD3 = "aimnet2_wb97md3"

    ## MACE
    MACE_MP_0 = "mace_mp_0"
    MACE_MP_0B2_L = "mace_mp_0b2_l"

    OMOL25_CONSERVING_S = "omol25_conserving_s"

    ## UMA
    UMA_S_OMOL = "uma_s_omol"
    UMA_M_OMOL = "uma_m_omol"
    UMA_S_OMAT = "uma_s_omat"
    UMA_M_OMAT = "uma_m_omat"
    UMA_S_OMC = "uma_s_omc"
    UMA_M_OMC = "uma_m_omc"

    ## Orb
    ORB_V3_CONSERVATIVE_INF_OMAT = "orb_v3_conservative_inf_omat"
    ORB_V3_CONSERVATIVE_OMOL = "orb_v3_conservative_omol"

    ## Egret
    EGRET_1 = "egret_1"
    EGRET_1E = "egret_1e"
    EGRET_1T = "egret_1t"

    # Semiempirical methods
    GFN_FF = "gfn_ff"
    GFN0_XTB = "gfn0_xtb"
    GFN1_XTB = "gfn1_xtb"
    GFN2_XTB = "gfn2_xtb"
    G_XTB = "g_xtb"

    # Force fields
    OFF_SAGE_2_0_0 = "off_sage_2_0_0"
    OFF_SAGE_2_2_1 = "off_sage_2_2_1"

    # Deprecated force fields
    SMIRNOFF_2_0_0_AMBER_AM1BCC = "smirnoff_2_0_0_amber_am1bcc"  # Deprecated: use OFF_SAGE_2_0_0
    SMIRNOFF_2_2_1_AMBER_AM1BCC = "smirnoff_2_2_1_amber_am1bcc"  # Deprecated: use OFF_SAGE_2_2_1

    def default_engine(self, *, is_periodic: bool = False) -> Engine:
        """
        Return the canonical Engine for this quantum-chemistry method.

        :param is_periodic: if True and method is in XTB family, return
            "tblite" (periodic-capable backend) instead of "xtb"
        :return: lower-case engine identifier (e.g. "psi4", "mace")

        >>> Method.MACE_MP_0B2_L.default_engine().value
        'mace'
        >>> Method.GFN2_XTB.default_engine().value
        'xtb'
        >>> Method.GFN2_XTB.default_engine(is_periodic=True).value
        'tblite'
        """
        match self:
            case Method.AIMNET2_WB97MD3:
                return Engine.AIMNET2
            case Method.MACE_MP_0B2_L:
                return Engine.MACE
            case (
                Method.OMOL25_CONSERVING_S | Method.UMA_S_OMOL | Method.UMA_M_OMOL | Method.UMA_S_OMAT | Method.UMA_M_OMAT | Method.UMA_S_OMC | Method.UMA_M_OMC
            ):
                return Engine.OMOL25
            case Method.ORB_V3_CONSERVATIVE_INF_OMAT | Method.ORB_V3_CONSERVATIVE_OMOL:
                return Engine.ORB
            case method if method in XTB_METHODS:
                return Engine.TBLITE if is_periodic else Engine.XTB
            case Method.OFF_SAGE_2_0_0 | Method.OFF_SAGE_2_2_1 | Method.SMIRNOFF_2_0_0_AMBER_AM1BCC | Method.SMIRNOFF_2_2_1_AMBER_AM1BCC:
                return Engine.OPENFF
            case Method.EGRET_1 | Method.EGRET_1E | Method.EGRET_1T:
                return Engine.EGRET
            case Method.SKALA:
                return Engine.PYSCF
            case _:
                # All remaining methods (HF, DFT, composite, etc.) fall back to Psi4
                return Engine.PSI4


PrepackagedNNPMethod = Literal[
    Method.AIMNET2_WB97MD3,
    Method.OMOL25_CONSERVING_S,
    Method.UMA_S_OMOL,
    Method.UMA_M_OMOL,
    Method.UMA_S_OMAT,
    Method.UMA_M_OMAT,
    Method.UMA_S_OMC,
    Method.UMA_M_OMC,
    Method.ORB_V3_CONSERVATIVE_INF_OMAT,
    Method.ORB_V3_CONSERVATIVE_OMOL,
    Method.EGRET_1,
    Method.EGRET_1E,
    Method.EGRET_1T,
]

PREPACKAGED_NNP_METHODS = {
    Method.AIMNET2_WB97MD3,
    Method.OMOL25_CONSERVING_S,
    Method.UMA_S_OMOL,
    Method.UMA_M_OMOL,
    Method.UMA_S_OMAT,
    Method.UMA_M_OMAT,
    Method.UMA_S_OMC,
    Method.UMA_M_OMC,
    Method.ORB_V3_CONSERVATIVE_INF_OMAT,
    Method.ORB_V3_CONSERVATIVE_OMOL,
    Method.EGRET_1,
    Method.EGRET_1E,
    Method.EGRET_1T,
}

CorrectableNNPMethod = Literal[Method.MACE_MP_0, Method.MACE_MP_0B2_L]
CORRECTABLE_NNP_METHODS = {Method.MACE_MP_0, Method.MACE_MP_0B2_L}

NNPMethod = PrepackagedNNPMethod | CorrectableNNPMethod
NNP_METHODS = PREPACKAGED_NNP_METHODS | CORRECTABLE_NNP_METHODS

XTBMethod = Literal[Method.GFN_FF, Method.GFN0_XTB, Method.GFN1_XTB, Method.GFN2_XTB, Method.G_XTB]
XTB_METHODS = {Method.GFN_FF, Method.GFN0_XTB, Method.GFN1_XTB, Method.GFN2_XTB, Method.G_XTB}

CompositeMethod = Literal[Method.HF3C, Method.B973C, Method.R2SCAN3C, Method.WB97X3C]
COMPOSITE_METHODS = {Method.HF3C, Method.B973C, Method.R2SCAN3C, Method.WB97X3C}

FFMethod = Literal[Method.OFF_SAGE_2_0_0, Method.OFF_SAGE_2_2_1]
FF_METHODS = {Method.OFF_SAGE_2_0_0, Method.OFF_SAGE_2_2_1}

PrepackagedMethod = XTBMethod | CompositeMethod | PrepackagedNNPMethod | FFMethod
PREPACKAGED_METHODS = XTB_METHODS | COMPOSITE_METHODS | PREPACKAGED_NNP_METHODS | FF_METHODS

MethodWithCorrection = Literal[Method.WB97XD3, Method.WB97XV, Method.WB97MV, Method.WB97MD3BJ, Method.DSDBLYPD3BJ]
METHODS_WITH_CORRECTION = {Method.WB97XD3, Method.WB97XV, Method.WB97MV, Method.WB97MD3BJ, Method.DSDBLYPD3BJ, Method.B97D3BJ}

MGGAFunctional = Literal[Method.R2SCAN, Method.R2SCAN3C, Method.TPSS, Method.TPSSH, Method.M06L, Method.M06, Method.M062X, Method.WB97MD3BJ, Method.WB97MV]
MGGA_FUNCTIONALS = {Method.R2SCAN, Method.R2SCAN3C, Method.TPSS, Method.TPSSH, Method.M06L, Method.M06, Method.M062X, Method.WB97MD3BJ, Method.WB97MV}

RangeSeparatedFunctional = Literal[Method.CAMB3LYP, Method.WB97XD3, Method.WB97XV, Method.WB97X3C, Method.WB97MV, Method.WB97MD3BJ]
RANGE_SEPARATED_FUNCTIONALS = {Method.CAMB3LYP, Method.WB97XD3, Method.WB97XV, Method.WB97X3C, Method.WB97MV, Method.WB97MD3BJ}

DFTFunctional = Literal[
    Method.PBE,
    Method.PBE0,
    Method.BP86,
    Method.B3LYP,
    Method.B973C,
    Method.B97D3BJ,
    Method.R2SCAN,
    Method.R2SCAN3C,
    Method.TPSS,
    Method.TPSSH,
    Method.M06L,
    Method.M06,
    Method.M062X,
    Method.CAMB3LYP,
    Method.WB97XD3,
    Method.WB97XV,
    Method.WB97X3C,
    Method.WB97MV,
    Method.WB97MD3BJ,
    Method.DSDBLYPD3BJ,
    Method.SKALA,
]
DFT_FUNCTIONALS = {
    Method.PBE,
    Method.PBE0,
    Method.BP86,
    Method.B3LYP,
    Method.B973C,
    Method.B97D3BJ,
    Method.R2SCAN,
    Method.R2SCAN3C,
    Method.TPSS,
    Method.TPSSH,
    Method.M06L,
    Method.M06,
    Method.M062X,
    Method.CAMB3LYP,
    Method.WB97XD3,
    Method.WB97XV,
    Method.WB97X3C,
    Method.WB97MV,
    Method.WB97MD3BJ,
    Method.DSDBLYPD3BJ,
    Method.SKALA,
}



"""Contains functions for dealing with the .cif file format."""

import re
from collections import deque
from datetime import datetime
from typing import Any, Callable

import numpy as np

from .data import CODES


def mmcif_string_to_mmcif_dict(filestring: str) -> dict[str, Any]:
    """
    Converts a .cif filestring into a dict that represents its table structure.
    Only lines which aren't empty and which don't begin with # are used.

    Multi-line strings are consolidated onto one line, and the whole thing is
    then split into the blocks that will become table lists. At the end, quote
    marks are removed from any string which retains them.

    :param filestring: .cif filestring to process
    """

    lines = deque(filter(lambda l: l and l[0] != "#", filestring.split("\n")))
    lines = consolidate_strings(lines)
    blocks = mmcif_lines_to_mmcif_blocks(lines)

    mmcif_dict = {
        block["category"]: loop_block_to_list(block)
        if block["lines"][0] == "loop_"  # keep open
        else non_loop_block_to_list(block)
        for block in blocks
    }
    strip_quotes(mmcif_dict)

    return mmcif_dict


def consolidate_strings(lines: deque[str]) -> deque[str]:
    """
    Generally, .cif files have a one file line to one table row correspondence.
    Sometimes however, a string cell is given a line of its own, breaking the
    row over several lines. This function takes the lines of a .cif file and
    puts all table rows on a single line.

    :param lines: .cif file lines
    :return: consolidated .cif file lines
    """

    new_lines: deque[str] = deque()
    while lines:
        line = lines.popleft()
        if line.startswith(";"):
            string = [line[1:].strip()]
            while not lines[0].startswith(";"):
                string.append(lines.popleft())
            lines.popleft()
            new_lines[-1] += ' "{}"'.format(" ".join(string).replace('"', "\x1a").replace("'", "\x1b"))
        else:
            new_lines.append(line)
    return new_lines


def mmcif_lines_to_mmcif_blocks(lines: deque[str]) -> list[dict[str, Any]]:
    """
    Takes a list of .cif file lines and splits them into table blocks. Each
    block will be a dict containing a category name and a list of lines.

    :param lines: .cif file lines
    """

    category = None
    block: list[str] = []
    blocks = []
    while lines:
        line = lines.popleft()
        if line.startswith("data_"):
            continue
        if line.startswith("_"):
            line_category = line.split(".")[0]
            if line_category != category:
                if category:
                    blocks.append({"category": category[1:], "lines": block})
                category = line_category
                block = []
        if line.startswith("loop_"):
            if category:
                blocks.append({"category": category[1:], "lines": block})
            category = lines[0].split(".")[0]
            block = []
        block.append(line)
    if block:
        blocks.append({"category": category[1:], "lines": block})  # type: ignore [index]
    return blocks


def non_loop_block_to_list(block: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Takes a simple block dict with no loop and turns it into a table list.

    :param block: .cif block to process
    :return: processed table list
    """

    d = {}
    for index in range(len(block["lines"]) - 1):
        if block["lines"][index + 1][0] != "_":
            block["lines"][index] += " " + block["lines"][index + 1]
    block["lines"] = [l for l in block["lines"] if l[0] == "_"]
    for line in block["lines"]:
        name = line.split(".")[1].split()[0]
        value = line
        if line.startswith("_"):
            value = " ".join(line.split()[1:])
        d[name] = value
    return [d]


def loop_block_to_list(block: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Takes a loop block dict where the initial lines are table headers and turns
    it into a table list. Sometimes a row is broken over several lines so this
    function deals with that too.

    :param block: .cif block to process
    :return: processed table list
    """

    names, lines, _ = [], [], True
    body_start = 0
    for index, line in enumerate(block["lines"][1:], start=1):
        if not line.startswith("_" + block["category"]):
            body_start = index
            break
    names = [l.split(".")[1].rstrip() for l in block["lines"][1:body_start]]
    lines = [split_values(l) for l in block["lines"][body_start:]]
    l = []
    for n in range(len(lines) - 1):
        while n < len(lines) - 1 and len(lines[n]) + len(lines[n + 1]) <= len(names):
            lines[n] += lines[n + 1]
            lines.pop(n + 1)
    for line in lines:
        l.append(dict(zip(names, line, strict=True)))

    return l


def split_values(line: str) -> list[str]:
    """
    The body of a .cif table is a series of lines, with each cell divided by
    whitespace. This function takes a string line and breaks it into cells.

    There are a few peculiarities to handle. Sometimes a cell is a string
    enclosed in quote marks, and spaces within this string obviously shouldn't
    be used to break the line. This function handles all of that.

    :param line: .cif line to split
    :return: list of cell values
    """

    if not re.search("['\"]", line):
        return line.split()
    chars = deque(line.strip())
    values, in_string = [], False
    value: list[Any] = []
    while chars:
        char = chars.popleft()
        if char == " " and not in_string:
            values.append(value)
            value = []
        elif char in "'\"":
            if in_string and chars and chars[0] != " ":
                value.append(char)
            else:
                in_string = not in_string
        else:
            value.append(char)
    values.append(value)

    return ["".join(v) for v in values if v]


def strip_quotes(mmcif_dict: dict[str, Any]) -> None:
    """
    In-place removes unneeded quote marks from a .mmcif dictionary.

    :param mmcif_dict: almost finished .mmcif dictionary to clean
    """
    for _, table in mmcif_dict.items():
        for row in table:
            for k, value in row.items():
                for char in "'\"":
                    if value[0] == char and value[-1] == char:
                        row[k] = value[1:-1]
                    row[k] = row[k].replace("\x1a", '"').replace("\x1b", "'")


def mmcif_dict_to_data_dict(mmcif_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Converts an .mmcif dictionary into an atomium data dictionary, with the
    same standard layout that the other file formats get converted into.

    :param mmcif_dict: .mmcif dictionary
    :return: atomium data dictionary
    """

    data_dict = {
        "description": {"code": None, "title": None, "deposition_date": None, "classification": None, "keywords": [], "authors": []},
        "experiment": {"technique": None, "source_organism": None, "expression_system": None, "missing_residues": []},
        "quality": {"resolution": None, "rvalue": None, "rfree": None},
        "geometry": {"assemblies": [], "crystallography": {}},
        "models": [],
    }
    update_description_dict(mmcif_dict, data_dict)
    update_experiment_dict(mmcif_dict, data_dict)
    update_quality_dict(mmcif_dict, data_dict)
    update_geometry_dict(mmcif_dict, data_dict)
    update_models_list(mmcif_dict, data_dict)

    return data_dict


def update_description_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its description sub-dictionary with
    information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "code", "entry", "id")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "title", "struct", "title")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "deposition_date", "pdbx_database_status", "recvd_initial_deposition_date", date=True)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "classification", "struct_keywords", "pdbx_keywords")
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "keywords", "struct_keywords", "text", split=True)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "description", "authors", "audit_author", "name", multi=True)


def update_experiment_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its experiment sub-dictionary with
    information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "technique", "exptl", "method")
    for cat, key in [
        ["entity_src_nat", "pdbx_organism_scientific"],
        ["entity_src_gen", "pdbx_gene_src_scientific_name"],
        ["pdbx_entity_src_syn", "organism_scientific"],
    ]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "source_organism", cat, key)
        if data_dict["experiment"]["source_organism"] not in [None, "?"]:
            break
    mmcif_to_data_transfer(mmcif_dict, data_dict, "experiment", "expression_system", "entity_src_gen", "pdbx_host_org_scientific_name")
    for r in mmcif_dict.get("pdbx_unobs_or_zero_occ_residues", []):
        insert = "" if r["PDB_ins_code"] in "?." else r["PDB_ins_code"]
        data_dict["experiment"]["missing_residues"].append({"id": f"{r['auth_asym_id']}.{r['auth_seq_id']}{insert}", "name": r["auth_comp_id"]})


def update_quality_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its quality sub-dictionary with
    information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "resolution", "reflns", "d_resolution_high", func=float)
    if not data_dict["quality"]["resolution"]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "resolution", "refine", "ls_d_res_high", func=float)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rvalue", "refine", "ls_R_factor_R_work", func=float)
    if not data_dict["quality"]["rvalue"]:
        mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rvalue", "refine", "ls_R_factor_obs", func=float)
    mmcif_to_data_transfer(mmcif_dict, data_dict, "quality", "rfree", "refine", "ls_R_factor_R_free", func=float)


def update_geometry_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its geometry sub-dictionary with
    information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    data_dict["geometry"]["assemblies"] = [
        {
            "id": int(a["id"]),
            "software": a.get("method_details", None),
            "delta_energy": None,
            "buried_surface_area": None,
            "surface_area": None,
            "transformations": [],
        }
        for a in mmcif_dict.get("pdbx_struct_assembly", [])
    ]
    operations = {
        o["id"]: [[float(o["matrix[{}][{}]".format(r, c)]) for c in [1, 2, 3]] + [float(o["vector[{}]".format(r)])] for r in [1, 2, 3]] + [[0, 0, 0, 1]]
        for o in mmcif_dict.get("pdbx_struct_oper_list", [])
    }
    for assembly in data_dict["geometry"]["assemblies"]:
        if assembly["software"] == "?":
            assembly["software"] = None
        assign_metrics_to_assembly(mmcif_dict, assembly)
        assign_transformations_to_assembly(mmcif_dict, operations, assembly)
    update_crystallography_dict(mmcif_dict, data_dict)


def assign_metrics_to_assembly(mmcif_dict: dict[str, Any], assembly: dict[str, Any]) -> None:
    """
    Takes an assembly dict, and goes through an mmcif dictionary looking for
    relevant energy etc. information to update it with.

    :param mmcif_dict: dictionary to read
    :param assembly: assembly to update
    """

    for a in mmcif_dict.get("pdbx_struct_assembly_prop", []):
        if a["biol_id"] == str(assembly["id"]):
            if a["type"] == "MORE":
                assembly["delta_energy"] = float(a["value"].split("/")[0])
            elif a["type"] == "SSA (A^2)":
                assembly["surface_area"] = float(a["value"].split("/")[0])
            elif a["type"] == "ABSA (A^2)":
                assembly["buried_surface_area"] = float(a["value"].split("/")[0])


def assign_transformations_to_assembly(mmcif_dict: dict[str, Any], operations: Any, assembly: dict[str, Any]) -> None:
    """
    Takes an assembly dict, and goes through an mmcif dictionary looking for
    relevant transformation information to update it with.

    :param mmcif_dict: .mmcif dictionary to read
    :param operations: processed operations matrices
    :param assembly: assembly to update
    """

    for gen in mmcif_dict.get("pdbx_struct_assembly_gen", []):
        if gen["assembly_id"] == str(assembly["id"]):
            op_ids_groups = get_operation_id_groups(gen["oper_expression"])
            ops = operation_id_groups_to_operations(operations, op_ids_groups)
            for operation in ops:
                assembly["transformations"].append(
                    {"chains": gen["asym_id_list"].split(","), "matrix": [row[:3] for row in operation[:3]], "vector": [row[-1] for row in operation[:3]]}
                )


def get_operation_id_groups(expression: str) -> list[list[str]]:
    """
    Determines which transformation IDs are an operator expression is referring to.

    For example, (1,2,3) becomes [[1, 2, 3]], (1-3)(8-11,17) becomes
        [[1, 2, 3], [8, 9, 10, 11, 17]], and so on.

    :param expression: expression to parse
    :return: list of transformation ID groups
    """
    if expression[0] != "(":
        expression = "({})".format(expression)

    group_ids = []
    for group in re.findall(r"\((.+?)\)", expression):
        ids = []
        elements = group.split(",")
        for element in elements:
            if "-" in element:
                bounds = [int(x) for x in element.split("-")]
                ids += [str(n) for n in list(range(bounds[0], bounds[1] + 1))]
            else:
                ids.append(element)

        group_ids.append(ids)

    return group_ids


def update_crystallography_dict(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its crystallography sub-sub-dictionary
    with information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    if mmcif_dict.get("cell"):
        mmcif_to_data_transfer(mmcif_dict, data_dict["geometry"], "crystallography", "space_group", "symmetry", "space_group_name_H-M")
        data_dict["geometry"]["crystallography"]["unit_cell"] = [
            float(mmcif_dict["cell"][0][key].replace("?", "0")) for key in ["length_a", "length_b", "length_c", "angle_alpha", "angle_beta", "angle_gamma"]
        ]
    if data_dict["geometry"]["crystallography"].get("space_group") == "NA":
        data_dict["geometry"]["crystallography"] = {}


def operation_id_groups_to_operations(operations: Any, operation_id_groups: Any) -> Any:
    """
    Creates a operation matrices for an assembly, from operation IDs - cross
    multiplying as required.

    :param operations: parsed .mmcif operations
    :param operation_id_groups: operation IDs
    :return: operation matrices
    """
    operation_groups = [[operations[i] for i in ids] for ids in operation_id_groups]

    while operation_groups and len(operation_groups) != 1:
        operations = [np.matmul(op1, op2) for op1 in operation_groups[0] for op2 in operation_groups[1]]
        operation_groups[0] = operations
        operation_groups.pop(1)

    return operation_groups[0]


def update_models_list(mmcif_dict: dict[str, Any], data_dict: dict[str, Any]) -> None:
    """
    Takes a data dictionary and updates its models list with
    information from a .mmcif dictionary.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    """

    data_dict["models"] = []
    types = {e["id"]: e["type"] for e in mmcif_dict.get("entity", {})}
    names = {e["id"]: e["name"] for e in mmcif_dict.get("chem_comp", {}) if e.get("mon_nstd_flag", "n") != "y"}
    entities = {m["id"]: m["entity_id"] for m in mmcif_dict.get("struct_asym", [])}

    # sequences = make_sequences(mmcif_dict)
    secondary_structure = make_secondary_structure(mmcif_dict)
    aniso = make_aniso(mmcif_dict)

    model: dict[str, Any] = {"polymer": {}, "non_polymer": {}, "water": {}, "branched": {}}
    model_num = mmcif_dict["atom_site"][0]["pdbx_PDB_model_num"]
    for atom in mmcif_dict["atom_site"]:
        if atom["pdbx_PDB_model_num"] != model_num:
            data_dict["models"].append(model)
            model = {"polymer": {}, "non_polymer": {}, "water": {}, "branched": {}}
            model_num = atom["pdbx_PDB_model_num"]
        mol_type = types[entities[atom["label_asym_id"]]]
        if mol_type in {"polymer", "branched"}:
            add_atom_to_polymer(atom, aniso, model, names)
        else:
            add_atom_to_non_polymer(atom, aniso, model, mol_type, names)
    data_dict["models"].append(model)

    for model in data_dict["models"]:
        add_sequences_to_polymers(model, mmcif_dict, entities)
        add_secondary_structure_to_polymers(model, secondary_structure)


def make_aniso(mmcif_dict: dict[str, Any]) -> dict[int, list[str]]:
    """
    Makes a mapping of atom IDs to anisotropy information.

    :param mmcif_dict: .mmcif dict to read
    :return: anisotropy mapping
    """
    return {
        int(a["id"]): [float(a["U[{}][{}]".format(x, y)]) for x, y in ["11", "22", "33", "12", "13", "23"]]  # type: ignore [has-type, misc]
        for a in mmcif_dict.get("atom_site_anisotrop", [])
    }


def make_secondary_structure(mmcif_dict: dict[str, Any]) -> dict[str, list[list[str]]]:
    """
    Creates a dictionary of helices and strands, with each having a list of
    start and end residues.

    :param mmcif_dict: .mmcif dict to read
    :return: secondary structure dictionary
    """
    helices = [
        [
            "{}.{}{}".format(
                helix[f"{x}_auth_asym_id"],
                helix[f"{x}_auth_seq_id"],
                helix[f"pdbx_{x}_PDB_ins_code"].replace("?", ""),
            )
            for x in ["beg", "end"]
        ]
        for helix in mmcif_dict.get("struct_conf", [])
    ]

    strands = [
        [
            "{}.{}{}".format(
                strand[f"{x}_auth_asym_id"],
                strand[f"{x}_auth_seq_id"],
                strand[f"pdbx_{x}_PDB_ins_code"].replace("?", ""),
            )
            for x in ["beg", "end"]
        ]
        for strand in mmcif_dict.get("struct_sheet_range", [])
    ]
    return {"helices": helices, "strands": strands}


def add_atom_to_polymer(atom: dict[str, Any], aniso: dict[int, Any], model: dict[str, Any], names: dict[str, Any]) -> None:
    """
    Takes an MMCIF atom dictionary, converts it, and adds it to a polymer dictionary.

    :param atom: .mmcif dictionary to read
    :param aniso: lookup dictionary for anisotropy information
    :param model: model to update
    :param names: lookup dictionary for full name information
    """

    mol_id = atom["auth_asym_id"]
    res_id = make_residue_id(atom)
    try:
        model["polymer"][mol_id]["residues"][res_id]["atoms"][int(atom["id"])] = atom_dict_to_atom_dict(atom, aniso)
    except Exception:
        name = atom.get("auth_comp_id") or atom.get("label_comp_id") or "UNKNOWN"
        try:
            model["polymer"][mol_id]["residues"][res_id] = {
                "name": name,
                "full_name": names.get(name).upper() if names.get(name) is not None else None,  # type: ignore [union-attr]
                "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
                "number": len(model["polymer"][mol_id]["residues"]) + 1,
            }
        except Exception:
            model["polymer"][mol_id] = {
                "internal_id": atom["label_asym_id"],
                "helices": [],
                "strands": [],
                "residues": {
                    res_id: {
                        "name": name,
                        "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
                        "number": 1,
                        "full_name": names.get(name).upper() if names.get(name) is not None else None,  # type: ignore [union-attr]
                    }
                },
            }


def add_atom_to_non_polymer(atom: dict[str, Any], aniso: dict[int, Any], model: dict[str, Any], mol_type: str, names: dict[str, Any]) -> None:
    """
    Takes an MMCIF atom dictionary, converts it, and adds it to a non_polymer dictionary.

    :param atom: .mmcif dictionary to read
    :param aniso: lookup dictionary for anisotropy information
    :param model: model to update
    :param mol_type: non_polymer or water
    :param names: lookup dictionary for full name information
    """
    mol_id = make_residue_id(atom)
    if mol_type == "non-polymer":
        mol_type = "non_polymer"

    try:
        model[mol_type][mol_id]["atoms"][int(atom["id"])] = atom_dict_to_atom_dict(atom, aniso)
    except Exception:
        name = atom.get("auth_comp_id") or atom.get("label_comp_id") or "UNKNOWN"
        model[mol_type][mol_id] = {
            "name": name,
            "full_name": names.get(name).upper() if names.get(name) is not None and names.get(name).lower() != "water" else None,  # type: ignore [union-attr]
            "internal_id": atom["label_asym_id"],
            "polymer": atom["auth_asym_id"],
            "atoms": {int(atom["id"]): atom_dict_to_atom_dict(atom, aniso)},
        }


def make_residue_id(d: dict[str, Any]) -> str:
    """
    Generates a residue ID for an atom.

    :param d: atom dictionary to read
    :return: residue ID
    """
    insert = "" if d["pdbx_PDB_ins_code"] in "?." else d["pdbx_PDB_ins_code"]

    return "{}.{}{}".format(d["auth_asym_id"], d["auth_seq_id"], insert)


def add_sequences_to_polymers(model: dict[str, Any], mmcif_dict: dict[str, Any], entities: dict[str, Any]) -> None:
    """
    Takes a pre-populated mapping of chain IDs to entity IDs, and uses them
    to add sequence information to a model.

    :param model: model to update
    :param mmcif_dict: .mmcif dictionary to read
    :param entities: mapping of chain IDs to entity IDs
    """
    sequences = make_sequences(mmcif_dict)
    for polymer in model["polymer"].values():
        polymer["sequence"] = sequences.get(entities.get(polymer["internal_id"], ""), "")


def add_secondary_structure_to_polymers(model: dict[str, Any], ss_dict: dict[str, Any]) -> None:
    """
    Updates polymer dictionaries with secondary structure information, from
    a previously created mapping.

    :param model: model to update
    :param ss_dict: mapping to read
    """
    for ss in ("helices", "strands"):
        for segment in ss_dict[ss]:
            chain = model["polymer"].get(segment[0].split(".")[0])
            if chain:
                in_segment = False
                chain[ss].append([])
                for residue_id in chain["residues"].keys():
                    if residue_id == segment[0]:
                        in_segment = True
                    if in_segment:
                        chain[ss][-1].append(residue_id)
                    if residue_id == segment[1]:
                        break


def make_sequences(mmcif_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Creates a mapping of entity IDs to sequences.

    :param mmcif_dict: .mmcif dictionary to read
    :return: sequence mapping
    """
    return {
        e["id"]: "".join([CODES.get(res["mon_id"], "X") for res in mmcif_dict.get("entity_poly_seq", []) if res["entity_id"] == e["id"]])
        for e in mmcif_dict.get("entity", [])
        if e["type"] == "polymer"
    }


def atom_dict_to_atom_dict(d: dict[str, Any], aniso_dict: dict[int, Any]) -> dict[str, Any]:
    """
    Turns an .mmcif atom dictionary into an atomium atom data dictionary.

    :param d: .mmcif atom dictionary
    :param aniso_dict: mapping of atom IDs to anisotropy
    :return: atom data dictionary
    """
    charge = "pdbx_formal_charge"

    atom = {
        "x": d["Cartn_x"],
        "y": d["Cartn_y"],
        "z": d["Cartn_z"],
        "element": d["type_symbol"],
        "name": d.get("label_atom_id"),
        "occupancy": d.get("occupancy", 1),
        "bvalue": d.get("B_iso_or_equiv"),
        "charge": d.get(charge, 0) if d.get(charge) != "?" else 0,
        "alt_loc": d.get("label_alt_id") if d.get("label_alt_id") != "." else None,
        "anisotropy": aniso_dict.get(int(d["id"]), None),
        "is_hetatm": d.get("group_PDB", "ATOM") == "HETATM",
    }

    for key in ["x", "y", "z", "charge", "bvalue", "occupancy"]:
        if atom[key] is not None:
            atom[key] = float(atom[key])

    if atom["charge"] == 0:
        atom["charge"] = None
    if not atom["is_hetatm"]:
        atom["is_hetatm"] = None
    if not atom["alt_loc"]:
        atom["alt_loc"] = None
    if atom["occupancy"] == 1:
        atom["occupancy"] = None
    if atom["name"] == atom["element"]:
        atom["name"] = None

    return atom


def mmcif_to_data_transfer(
    mmcif_dict: dict[str, Any],
    data_dict: dict[str, Any],
    d_cat: Any,
    d_key: Any,
    m_table: Any,
    m_key: Any,
    date: bool = False,
    split: bool = False,
    multi: bool = False,
    func: Callable[[Any], Any] | None = None,
) -> None:
    """
    Function for transfering a bit of data from a .mmcif dictionary to a
    data dictionary, or doing nothing if the data doesn't exist.

    :param mmcif_dict: .mmcif dictionary to read
    :param data_dict: data dictionary to update
    :param d_cat: top-level key in the data dictionary
    :param d_key: data dictionary field to update
    :param m_table: name of the .mmcif table to look in
    :param m_key: .mmcif field to read
    :param date: if True, value will be converted to a date
    :param split: if True, value will be split on commas
    :param multi: if True, every row in the table will be read
    :param func: if given, will be applied to the value
    """
    try:
        if multi:
            value = [row[m_key] for row in mmcif_dict[m_table]]
        else:
            value = mmcif_dict[m_table][0][m_key]

        if date:
            value = datetime.strptime(value, "%Y-%m-%d").date()  # type: ignore [arg-type, assignment]
        if split:
            value = value.replace(", ", ",").split(",")  # type: ignore [attr-defined]
        if func:
            value = func(value)

        if isinstance(value, str) and value == "?":
            value = None

        data_dict[d_cat][d_key] = value

    except Exception:
        pass



from typing import Any, Self, TypeVar

from pydantic import PositiveFloat, computed_field, field_validator, model_validator

from .base import Base, LowercaseStrEnum, UniqueList
from .basis_set import BasisSet
from .compute_settings import ComputeSettings
from .correction import Correction
from .engine import Engine
from .excited_state_settings import ExcitedStateSettingsUnion
from .method import CORRECTABLE_NNP_METHODS, DFT_FUNCTIONALS, METHODS_WITH_CORRECTION, PREPACKAGED_METHODS, RANGE_SEPARATED_FUNCTIONALS, Method
from .mode import Mode
from .opt_settings import OptimizationSettings
from .scf_settings import SCFSettings
from .solvent import SolventModel, SolventSettings
from .task import Task
from .thermochem_settings import ThermochemistrySettings

_T = TypeVar("_T")


class OmegaTuning(LowercaseStrEnum):
    """Options for omega (range-separation parameter) tuning"""

    KOOPMANS = "koopmans"  # Baer et al. doi.org/10.1146/annurev.physchem.012809.103321


class Settings(Base):
    """
    Settings for (base) computational chemistry calculations.

    :param mode: calculation mode (affects thresholds for optimizations, etc.)
    :param tasks: tasks to perform (deprecated, specify in workflows or calculations instead)
    :param method: computational method
    :param basis_set: basis set to use (autopopulated for 3c methods)
    :param engine: computational engine to use (auto-selected if None)
    :param corrections: list of corrections to apply (e.g. D3BJ, D4)
    :param solvent_settings: solvent model settings (if any)
    :param omega: range-separation parameter (Bohr⁻¹) or method to tune it (optional)
    :param scf_settings: SCF settings
    :param opt_settings: geometry optimization settings
    :param thermochem_settings: thermochemistry settings
    :param compute_settings: hardware settings
    """

    mode: Mode = Mode.AUTO

    # DEPRECATED - specify tasks only in BasicCalculationWorkflow or Calculation now
    tasks: UniqueList[Task] = [Task.ENERGY, Task.CHARGE, Task.DIPOLE]

    method: Method = Method.HARTREE_FOCK
    basis_set: BasisSet | None = None
    engine: Engine = None  # type: ignore [assignment]
    corrections: UniqueList[Correction] = []
    solvent_settings: SolventSettings | None = None
    omega: OmegaTuning | PositiveFloat | None = None

    excited_state_settings: ExcitedStateSettingsUnion | None = None

    # scf/opt settings will be set automatically based on mode, but can be overridden manually
    scf_settings: SCFSettings = SCFSettings()
    opt_settings: OptimizationSettings = OptimizationSettings()
    thermochem_settings: ThermochemistrySettings = ThermochemistrySettings()
    compute_settings: ComputeSettings = ComputeSettings()

    @model_validator(mode="after")
    def set_engine(self) -> Self:
        """Set the calculation engine."""
        self.engine = self.engine or self.method.default_engine()
        return self

    # mypy has this dead wrong (https://docs.pydantic.dev/2.0/usage/computed_fields/)
    # Python 3.12 narrows the reason for the ignore to prop-decorator
    @computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def level_of_theory(self) -> str:
        corrections = list(filter(lambda x: x not in (None, ""), self.corrections))

        if self.method in CORRECTABLE_NNP_METHODS:
            method = self.method.value if not corrections else f"{self.method.value}-{'-'.join(c.value for c in corrections)}"
        elif self.method in PREPACKAGED_METHODS or self.basis_set is None:
            method = self.method.value
        elif self.method in METHODS_WITH_CORRECTION or not corrections:
            method = f"{self.method.value}/{self.basis_set.name.lower()}"
        else:
            method = f"{self.method.value}-{'-'.join(c.value for c in corrections)}/{self.basis_set.name.lower()}"

        if self.solvent_settings is not None:
            method += f"/{self.solvent_settings.model.value}({self.solvent_settings.solvent.value})"

        return method

    @field_validator("mode")
    @classmethod
    def set_mode_auto(cls, mode: Mode) -> Mode:
        """Set the mode to RAPID if AUTO is selected."""
        if mode == Mode.AUTO:
            return Mode.RAPID

        return mode

    @model_validator(mode="after")
    def validate_and_build(self) -> Self:
        if self.mode == Mode.AUTO:
            self.mode = Mode.RAPID

        self.opt_settings = _assign_opt_settings_by_mode(self.mode, self.opt_settings)

        if self.omega is not None and self.method not in RANGE_SEPARATED_FUNCTIONALS:
            functionals = "\n    ".join(RANGE_SEPARATED_FUNCTIONALS)
            raise ValueError(f"Omega tuning may only be specified for range-separated DFT functionals:\n    {functionals}.")

        if self.excited_state_settings:
            if self.engine not in {Engine.PYSCF, Engine.GPU4PYSCF}:
                raise ValueError("Excited-state calculations are only supported with the PySCF and GPU4PySCF engines.")

            if self.method not in DFT_FUNCTIONALS:
                functionals = "\n    ".join(DFT_FUNCTIONALS)
                raise ValueError(f"Excited-state calculations may only be performed with DFT:\n    {functionals}.")

        if self.solvent_settings:
            if self.solvent_settings.model == SolventModel.COSMO and self.engine == Engine.GPU4PYSCF:
                raise ValueError("GPU4PySCF does not support the COSMO solvent model")

        return self

    def model_post_init(self, __context: Any, /) -> None:
        # figure out `optimize_ts`
        if Task.OPTIMIZE_TS in self.tasks:
            self.tasks.pop(self.tasks.index(Task.OPTIMIZE_TS))
            self.tasks.append(Task.OPTIMIZE)
            self.opt_settings.transition_state = True

        # composite methods have their own basis sets, so overwrite user stuff
        if self.method == Method.HF3C:
            self.basis_set = BasisSet(name="minix")
        elif self.method == Method.B973C:
            self.basis_set = BasisSet(name="def2-mTZVP")
        elif self.method == Method.R2SCAN3C:
            self.basis_set = BasisSet(name="def2-mTZVPP")
        elif self.method == Method.WB97X3C:
            self.basis_set = BasisSet(name="vDZP")

    @field_validator("basis_set", mode="before")
    @classmethod
    def parse_basis_set(cls, v: Any) -> BasisSet | dict[str, Any] | None:
        """Turn a string into a BasisSet object. (This is a little crude.)"""
        if isinstance(v, BasisSet):
            return None if v.name is None else v
        elif isinstance(v, dict):
            return None if v.get("name") is None else v
        elif isinstance(v, str):
            if len(v):
                return BasisSet(name=v)
            # "" is basically None, let's be real here...
            return None
        elif v is None:
            return None
        else:
            raise ValueError(f"invalid value {v} for basis_set")

    @field_validator("corrections", mode="before")
    @classmethod
    def remove_empty_string(cls, v: list[_T]) -> list[_T]:
        """Remove empty string values."""
        return [c for c in v if c] if v is not None else v


def _assign_opt_settings_by_mode(mode: Mode, opt_settings: OptimizationSettings) -> OptimizationSettings:
    """
    Assign optimization settings based on the mode.

    Constraints lead to a lot of noise, so we need to loosen the thresholds.

    cf. DLFIND manual, and https://www.cup.uni-muenchen.de/ch/compchem/geom/basic.html
    and the discussion at https://geometric.readthedocs.io/en/latest/how-it-works.html
    in periodic systems, "normal" is 0.05 eV/Å ~= 2e-3 Hartree/Å, and "careful" is 0.01 ~= 4e-4

    Note: thresholds here are in units of Hartree/Å, not Hartree/Bohr as listed in many places.
    """
    opt_settings.energy_threshold = 1e-6
    match mode:
        case Mode.RECKLESS:
            opt_settings.energy_threshold = 2e-5
            opt_settings.max_gradient_threshold = 7e-3
            opt_settings.rms_gradient_threshold = 6e-3
        case Mode.RAPID:
            opt_settings.energy_threshold = 5e-5
            opt_settings.max_gradient_threshold = 5e-3
            opt_settings.rms_gradient_threshold = 3.5e-3
        case Mode.CAREFUL:
            opt_settings.max_gradient_threshold = 9e-4
            opt_settings.rms_gradient_threshold = 6e-4
        case Mode.METICULOUS:
            opt_settings.max_gradient_threshold = 3e-5
            opt_settings.rms_gradient_threshold = 2e-5
        case Mode.DEBUG:
            opt_settings.max_gradient_threshold = 4e-6
            opt_settings.rms_gradient_threshold = 2e-6
        case _:
            raise ValueError(f"Unknown mode {mode.value}!")

    return opt_settings



from typing import Annotated, Self, Sequence

from pydantic import AfterValidator, NonNegativeInt

from .base import Base, round_optional_float
from .data import ELEMENT_SYMBOL, SYMBOL_ELEMENT
from .types import Vector3D, round_vector3d


class Atom(Base):
    """
    A single atom.

    :param atomic_number: atomic number
    :param position: Cartesian coordinates, in Å
    :param mass: mass, in Daltons; if None, most common isotopologue will be used
    """

    atomic_number: NonNegativeInt
    position: Annotated[Vector3D, AfterValidator(round_vector3d(8))]
    mass: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    def __repr__(self) -> str:
        """
        >>> Atom(atomic_number=2, position=[0, 1, 2])
        Atom(2, [0.00000, 1.00000, 2.00000])
        """
        x, y, z = self.position
        if self.mass is not None:
            return f"Atom({self.atomic_number}, [{x:.5f}, {y:.5f}, {z:.5f}], mass={self.mass:.3f})"
        return f"Atom({self.atomic_number}, [{x:.5f}, {y:.5f}, {z:.5f}])"

    def __str__(self) -> str:
        """
        >>> str(Atom(atomic_number=2, position=[0, 1, 2]))
        'He    0.0000000000    1.0000000000    2.0000000000'
        """
        x, y, z = self.position
        return f"{self.atomic_symbol:2} {x:15.10f} {y:15.10f} {z:15.10f}"

    @property
    def atomic_symbol(self) -> str:
        """
        >>> Atom(atomic_number=2, position=[0, 1, 2]).atomic_symbol
        'He'
        """
        return ELEMENT_SYMBOL[self.atomic_number]

    def edited(self, atomic_number: int | None = None, position: Sequence[float] | None = None, mass: float | None = None) -> Self:
        """
        Create a new Atom with the specified changes.

        >>> a = Atom(atomic_number=2, position=[0, 1, 2])
        >>> a2 = a.edited(3)
        >>> a is a2
        False
        >>> a2
        Atom(3, [0.00000, 1.00000, 2.00000])
        """
        if atomic_number is None:
            atomic_number = self.atomic_number
        if position is None:
            position = list(self.position)
        if mass is None:
            mass = self.mass

        return self.__class__(atomic_number=atomic_number, position=position, mass=mass)

    @classmethod
    def from_xyz(cls: type[Self], xyz_line: str) -> Self:
        """
        >>> Atom.from_xyz("H 0 0 0")
        Atom(1, [0.00000, 0.00000, 0.00000])
        """
        name, *xyz = xyz_line.split()
        symbol = int(name) if name.isdigit() else SYMBOL_ELEMENT[name.title()]
        if not len(xyz) == 3:
            raise ValueError("XYZ file should have 3 coordinates per atom")
        return cls(atomic_number=symbol, position=xyz)



"""Protein-binder-design workflow."""

from enum import Enum
from typing import Annotated

from pydantic import AfterValidator

from ..base import Base, LowercaseStrEnum, round_optional_float
from ..protein import ProteinSequence
from ..types import ProteinUUID
from .workflow import Workflow


class BoltzGenSecondaryStructure(Base):
    """
    Represents the secondary structure assignments for a protein.

    :param id: Optional identifier for this secondary structure annotation.
    :param sheet: String encoding the residue indices comprising β-sheet structures
                  (e.g., "1,3..11" for residues 1, and 3 through 11).
    :param helix: String encoding residue indices comprising helices.
    :param loop: String encoding residue indices comprising loop or coil regions.
    """

    id: str | None = None
    sheet: str | None = None
    helix: str | None = None
    loop: str | None = None


class BoltzGenProteinEntity(Base):
    """
    Represents a protein entity, either a designed or natural sequence.

    :param id: Unique identifier for the protein.
    :param sequence: Protein sequence, may contain amino acids and numbers for designed regions.
    :param secondary_structure: Optional assigned secondary structure.
    :param cyclic: Whether the protein is cyclic (True/False). Optional.
    """

    id: str
    sequence: str  # can include amino acids as well as numbers for designed regions
    secondary_structure: BoltzGenSecondaryStructure | None = None
    # binding_types: BindingType | None = None - we may want to add this later but not used in examples.
    cyclic: bool | None = None


class BoltzGenLigandEntity(Base):
    """
    Represents a ligand entity (non-protein), such as a small molecule.

    :param id: Unique identifier for the ligand.
    :param smiles: SMILES string representation of the ligand.
    """

    id: str
    smiles: str
    # binding_types: str | None = None - we may want to add this later but not used in examples.


class BoltzGenRegionSelection(Base):
    """
    Defines a region of a protein chain by specifying its chain identifier and (optionally) residue indices.

    :param chain_id: Identifier for the protein chain (e.g., 'A', 'B', etc.).
    :param residue_indices: Residues to select, specified as a string in the format "5..7,13" or "5..15,50..".
    """

    chain_id: str | None = None
    residue_indices: str | None = None


class BoltzGenProximityRegionSelection(BoltzGenRegionSelection):
    """
    Defines a region of a protein chain based on spatial proximity to a selection of residues.

    Inherits:
        BoltzGenRegionSelection

    :param radius: Radius in angstroms (Å) used to select all residues within proximity to the specified region.
    """

    radius: int | None = None


class BoltzGenBindingType(Base):
    """
    Represents the binding interface specification for a given protein chain.

    :param chain_id: Identifier for the protein chain (e.g., 'A', 'B', etc.).
    :param binding: Residue indices or regions that are required to participate in binding
        (e.g., "5..7,13" or "all" for the whole chain).
    :param not_binding: Residue indices or regions that should explicitly not participate in binding
        (e.g., "5..7,13" or "all" for excluding the entire chain).
    """

    chain_id: str | None = None
    binding: str | None = None
    not_binding: str | None = None


class BoltzGenSecondaryStructureOptions(str, Enum):
    UNSPECIFIED = "UNSPECIFIED"
    LOOP = "LOOP"
    HELIX = "HELIX"
    SHEET = "SHEET"


class BoltzGenDesignInsertion(Base):
    """
    Represents an insertion site for protein design in a specific chain.

    :param chain_id: Identifier of the chain where the insertion occurs.
    :param residue_index: Position in the chain after which the insertion is to be made.
    :param number_of_residues: Number of residues to insert at the specified site (can be a string pattern).
    :param secondary_structure: Desired secondary structure type for the inserted residues
        ("UNSPECIFIED", "LOOP", "HELIX", or "SHEET"). Optional.
    """

    chain_id: str
    residue_index: int
    number_of_residues: str
    secondary_structure: BoltzGenSecondaryStructureOptions | None = None


class BoltzGenFileEntity(Base):
    """
    Represents a protein structure input and its associated region selection and design specifications
    for the BoltzGen binder design workflow.

    :param uuid: Unique identifier for the protein structure.
    :param include: List of regions to include in the design or analysis context.
    :param exclude: List of regions to explicitly exclude from consideration (e.g., for ignoring noisy/irrelevant regions).
    :param include_proximity: List of regions defined by spatial proximity (e.g., residues within a given radius).
    :param binding_types: List of binding type constraints or permitted interface regions.
    :param design: List of regions that are being subject to design (mutable, allowed to change).
    :param secondary_structure: List of desired or annotated secondary structure definitions for selected regions.
    :param design_insertions: List of new regions to be inserted with specified properties (e.g., insertion sites, structure preferences).
    """

    uuid: ProteinUUID
    include: list[BoltzGenRegionSelection] = []
    exclude: list[BoltzGenRegionSelection] = []
    # fuse: None  - we may want to add this later but not used in examples.
    include_proximity: list[BoltzGenProximityRegionSelection] = []
    binding_types: list[BoltzGenBindingType] = []
    # structure_groups: None - we may want to add this later but not used in examples.
    design: list[BoltzGenRegionSelection] = []
    secondary_structure: list[BoltzGenSecondaryStructure] = []
    design_insertions: list[BoltzGenDesignInsertion] = []


class BoltzGenAtomSpecification(Base):
    """
    Atom specification for a protein chain, used for applying constraints or referencing atoms.

    :param chain_id: Identifier for the protein chain (e.g., "A", "B").
    :param index: Residue index the atom belongs to (integer, 1-based).
    :param atom_name: Name of the atom (e.g., "CA", "N", "O", etc.).
    """

    chain_id: str
    index: int
    atom_name: str


class BoltzGenConstraint(Base):
    """
    Describes a covalent or spatial constraint between two specified atoms in the context of protein design.

    :param atom1: First atom in the constraint.
    :param atom2: Second atom in the constraint.
    """

    atom1: BoltzGenAtomSpecification
    atom2: BoltzGenAtomSpecification


class BoltzGenInput(Base):
    """
    Represents the primary input schema for the boltzgen application.

    :param protein_entities: Protein chains that are designed or targeted for binding.
    :param ligand_entities: Small molecules or other non-protein ligands relevant to the design.
    :param file_entities: 3d protein structures and input settings related to them.
    :param constraints: Covalent bond constraints
    """

    protein_entities: list[BoltzGenProteinEntity] = []
    ligand_entities: list[BoltzGenLigandEntity] = []
    file_entities: list[BoltzGenFileEntity] = []
    constraints: list[BoltzGenConstraint] = []


class BoltzGenScores(Base):
    """
    Compact, interpretable metrics for a designed binder.
    ↑ higher is better, ↓ lower is better

    :param quality_score: aggregate model quality (↑)
    :param num_filters_passed: number of QC/heuristic filters passed (↑)
    :param iptm: inter-chain pTM confidence, 0–1 (↑)
    :param design_ptm: design pTM confidence, 0–1 (↑)
    :param min_interaction_pae: minimum interface PAE, in Å (↓)
    :param bb_rmsd: backbone RMSD, in Å (↓)
    :param delta_sasa_refolded: ΔSASA of interface after refolding, Å² (↑ typically indicates better burial)
    :param plip_hbonds_refolded: count of hydrogen bonds at the interface (↑)
    :param plip_saltbridge_refolded: count of salt bridges at the interface (↑)
    :param liability_score: composite liabilities score (↓)
    :param liability_high_severity_violations: count of high-severity liabilities (↓)
    :param liability_num_violations: total liability count (↓)
    :param helix: fraction helical content, 0–1
    :param sheet: fraction β-sheet content, 0–1
    :param loop: fraction loop/coil content, 0–1
    :param design_largest_hydrophobic_patch_refolded: largest hydrophobic patch area after refolding, Å²
    :param design_hydrophobicity: overall design hydrophobicity score (unitless)
    :param num_tokens: sequence length / token count
    """

    quality_score: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    num_filters_passed: int | None = None

    iptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_ptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    min_design_to_target_pae: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_to_target_iptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    min_interaction_pae: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    bb_rmsd: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    delta_sasa_refolded: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    plip_hbonds_refolded: int | None = None
    plip_saltbridge_refolded: int | None = None

    liability_score: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    liability_high_severity_violations: int | None = None
    liability_num_violations: int | None = None

    helix: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    sheet: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    loop: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_largest_hydrophobic_patch_refolded: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_hydrophobicity: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    num_tokens: int | None = None


class ProteinBinderDesignResult(Base):
    """
    The output; a designed binder.

    :param sequence: sequence
    :param bound_structure: PDB of structure bound to target
    :param scores: scores for generated structure
    """

    binder_sequence: str | None = None  # deprecated
    binder_sequences: list[ProteinSequence] | None = None
    bound_structure: ProteinUUID | None = None
    scores: BoltzGenScores | None = None


class BoltzGenProtocol(LowercaseStrEnum):
    """
    The predefined protocol used for generation + filtering.
    """

    PROTEIN_ANYTHING = "protein-anything"
    PEPTIDE_ANYTHING = "peptide-anything"
    PROTEIN_SMALL_MOLECULE = "protein-small_molecule"
    NANOBODY_ANYTHING = "nanobody-anything"


class BoltzGenSettings(Base):
    """
    The settings for running BoltzGen.

    :param num_designs: how many designs to generate
    :param protocol: which protocol to use
    :param binding_residue: dict mapping chain ID to which residues should bind;
        the string follows the BoltzGen format of specifying ranges of residue indices (refer to their documentation).
        examples include "5..7,13" or "5..15,50..".
    """

    protocol: BoltzGenProtocol = BoltzGenProtocol.PROTEIN_ANYTHING
    num_designs: int = 100
    budget: int = 20


class ProteinBinderDesignWorkflow(Workflow):
    """
    A workflow for generating proteins or peptides that bind to something.


    New:
    :param binder_design_input: input to protein binder design workflow
    :param binder_design_settings: settings for protein generation method employed
    :param generated_binders: output structures
    """

    binder_design_input: BoltzGenInput = BoltzGenInput()
    binder_design_settings: BoltzGenSettings = BoltzGenSettings()

    generated_binders: list[ProteinBinderDesignResult] = []



from typing import Annotated

from pydantic import AfterValidator

from ..base import Base, round_float
from ..conformers import ConformerClusteringSettings, ConformerGenSettingsUnion, ConformerProperties, ConformerSearchMixin, iMTDSettings
from ..method import Method
from ..mode import Mode
from ..settings import Settings
from ..solvent import Solvent, SolventModel, SolventSettings
from ..types import UUID
from .multistage_opt import MultiStageOptSettings
from .workflow import MoleculeWorkflow


class SolventDependentConformer(Base):
    """
    Stores a single conformer scored in many different solvents.

    :param calculation: conformer (as a calculation)
    :param free_energy_by_solvent: free energy in every solvent, in Hartree
    :param relative_free_energy_by_solvent: relative free energy vs. lowest-energy conformer in every solvent, in kcal/mol
    :param population_by_solvent: population in every solvent (number between 0 and 1)
    """

    calculation: UUID
    free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]
    relative_free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]
    population_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]


class SolventDependentConformersWorkflow(ConformerSearchMixin, MoleculeWorkflow):
    """
    Conformers are generated through `conf_gen_settings`.

    Clustering is then performed, and optimization is conducted.
    The optimization of the conformers is governed by `multistage_opt_settings`.

    Final conformer scoring is done through a multi-level scheme:
        - A single-point gas-phase energy run through `multistage_opt_settings.sp_settings`.
        - A separate thermal free-energy correction computed via a single-point Hessian (GFN2-xTB).
        - A per-solvent CPCM-X calculation.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param conf_gen_settings: settings for conformer generation
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)

    New:
    :param solvents: solvents to study
    :param conformers: output conformers with per-solvent energies and weights
    :param per_solvent_properties: metrics for how overall distribution changes by solvent
    :param relative_free_energy_by_solvent: how free energy changes by solvent, for predicting ∆G_transfer
    """

    solvents: list[Solvent] = [
        Solvent.HEXANE,
        Solvent.OCTANOL,
        Solvent.CHLOROFORM,
        Solvent.DIMETHYLSULFOXIDE,
        Solvent.WATER,
    ]

    conf_gen_settings: ConformerGenSettingsUnion | None = iMTDSettings(
        max_confs=None,
        speed="normal",
        solvent_settings=SolventSettings(solvent=Solvent.WATER, model=SolventModel.ALPB),
        reopt=False,
        mtd_method=Method.GFN_FF,
        energy_window=30,
    )

    conformer_clustering_settings: ConformerClusteringSettings | None = ConformerClusteringSettings()

    multistage_opt_settings: MultiStageOptSettings = MultiStageOptSettings(
        optimization_settings=[
            Settings(
                method="gfn2_xtb",
                solvent_settings=SolventSettings(
                    solvent=Solvent.WATER,
                    model=SolventModel.ALPB,
                ),
                tasks=["optimize"],
            )
        ],
        singlepoint_settings=Settings(method="g_xtb", tasks=["energy"]),
        mode=Mode.MANUAL,
    )

    conformers: list[SolventDependentConformer] = []
    per_solvent_properties: dict[Solvent, ConformerProperties] = {}
    relative_free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]] = {}



from ..base import LowercaseStrEnum
from .workflow import ProteinSequenceWorkflow


class MSAFormat(LowercaseStrEnum):
    """Format of the MSA."""

    COLABFOLD = "colabfold"
    CHAI = "chai"
    BOLTZ = "boltz"


class MSAWorkflow(ProteinSequenceWorkflow):
    """
    Workflow for generating a MSA from protein sequences.

    Inherited:
    :param initial_protein_sequences: protein sequences of interest

    New:
    :param output_formats: formats of MSA return files
    """

    output_formats: list[MSAFormat] = [MSAFormat.COLABFOLD]



from typing import Self

from pydantic import PositiveFloat, PositiveInt, model_validator

from .base import Base


class BasisSetOverride(Base):
    """
    Basis set override for specific atoms or elements.

    :param name: basis set name for override
    :param atomic_numbers: atomic numbers to override (mutually exclusive with atoms)
    :param atoms: 1-indexed atom indices to override (mutually exclusive with atomic_numbers)
    """

    name: str
    atomic_numbers: list[PositiveInt] | None = None
    atoms: list[PositiveInt] | None = None

    @model_validator(mode="after")
    def check_override(self) -> Self:
        # ^ is xor
        assert (self.atomic_numbers is not None) ^ (self.atoms is not None), "Exactly one of atomic_numbers or atoms must be specified!"
        return self


class BasisSet(Base):
    """
    Atomic orbital basis set.

    :param name: basis set name (e.g., def2-SVP, cc-pVDZ)
    :param overrides: element or atom-specific basis set overrides
    :param cutoff_threshold: basis function screening threshold
    """

    name: str

    # do we want to override the default basis set for specific atoms or elements?
    overrides: list[BasisSetOverride] | None = []

    # value below which a basis function can be ignored
    # (for improving DFT grid calcs, as per Stratmann/Scuseria/Frisch CPL 1996)
    # this shouldn't really need to be modified...
    cutoff_threshold: PositiveFloat = 1e-10



from .base import LowercaseStrEnum


class Mode(LowercaseStrEnum):
    """Calculation accuracy mode."""

    # choose based on job type
    AUTO = "auto"

    # very rapid, possibly unreliable
    RECKLESS = "reckless"

    # standard for single-point energies
    RAPID = "rapid"

    # standard for gradient / geometry optimizations
    CAREFUL = "careful"

    # very careful!
    METICULOUS = "meticulous"

    # for debugging purposes
    DEBUG = "debug"

    # set parameters manually, prevent overwriting
    MANUAL = "manual"



"""
Read the NIST isotopes data file and write it to a JSON file.

NIST Isotopes data from:
https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl?ele=&all=all&ascii=ascii2
"""

import json
from collections import defaultdict
from importlib import resources
from typing import Callable, TypeVar

data_dir = resources.files("stjames").joinpath("data")

_T = TypeVar("_T")


def process_line(line: str, fmt: Callable[[str], _T] = str) -> _T:  # type: ignore[assignment]
    """
    Process a line from the NIST data file.

    :param line: line to process
    :param fmt: function to format the value
    >>> process_line("Atomic Number = 1", int)
    1
    """
    return fmt(line.rsplit("=", maxsplit=1)[-1].strip())


def fmt_float(val: str) -> float:
    """
    Format a float from the NIST data file.

    >>> fmt_float(" 1.00784(7)")
    1.00784
    """
    return float(val.strip().split("(")[0])


def fmt_maybe_list(val: str) -> float:
    """
    Format a float or list of floats from the NIST data file.

    Only the first value is returned.

    >>> fmt_maybe_list("1.00784(7)")
    1.00784
    >>> fmt_maybe_list(" [1.00784,1.00811]")
    1.00784
    >>> fmt_maybe_list(" [98]")
    98.0
    """
    val = val.strip()
    if val.startswith("["):
        val = val[1:-1].split(",")[0]
    return fmt_float(val)


def process_chunk(chunk: str) -> tuple[int, int, tuple[float, float, float]]:
    r"""
    Atomic Number, Mass Number, (Relative Atomic Mass, Isotopic Composition, Standard Atomic Weight)

    >>> process_chunk('''\
    ... Atomic Number = 1
    ... Atomic Symbol = H
    ... Mass Number = 1
    ... Relative Atomic Mass = 1.00784(7)
    ... Isotopic Composition = 0.999885(70)
    ... Standard Atomic Weight = [1.00784,1.00811]
    ... Notes = m
    ... ''')
    (1, 1, (1.00784, 0.999885, 1.00784))
    """
    lines = chunk.splitlines()

    atomic_number = process_line(lines[0], int)
    _atomic_symbol = process_line(lines[1], str)
    mass_number = process_line(lines[2], int)
    relative_atomic_mass = process_line(lines[3], fmt_float)
    try:
        isotopic_composition = process_line(lines[4], fmt_float)
    except ValueError:
        isotopic_composition = 0
    try:
        standard_atomic_weight = process_line(lines[5], fmt_maybe_list)
    except ValueError:
        standard_atomic_weight = relative_atomic_mass

    return atomic_number, mass_number, (relative_atomic_mass, isotopic_composition, standard_atomic_weight)


def read_nist_isotopes() -> dict[int, dict[int, tuple[float, float, float]]]:
    """
    Read the NIST data file and write it to a JSON file.

    {Atomic Number: {Mass Number, (Relative Atomic Mass, Isotopic Composition, Standard Atomic Weight)}}
    """
    with data_dir.joinpath("nist_isotopes.txt").open() as f:
        next(f), next(f)  # Skip the first two lines
        nist_isotopes = f.read()

    isotopes: dict[int, dict[int, tuple[float, float, float]]] = defaultdict(dict)
    for chunk in nist_isotopes.split("\n\n"):
        atomic_number, mass_number, values = process_chunk(chunk)
        isotopes[atomic_number][mass_number] = values

    with open("nist_isotopes.json", "w") as f:
        json.dump(isotopes, f)

    return isotopes


if __name__ == "__main__":
    from pprint import pprint

    pprint(read_nist_isotopes())



"""Deprecated conformer search workflow, use ConformerSearchWorkflow instead."""

from typing import Annotated, Any

from pydantic import AfterValidator

from ..base import Base, round_float, round_optional_float
from ..constraint import Constraint
from ..method import Method
from ..mode import Mode
from ..solvent import Solvent
from .workflow import MoleculeWorkflow


class ConformerSettings(Base):
    num_confs_considered: int = 100
    num_confs_taken: int = 50
    rmsd_cutoff: float = 0.1

    transition_state: bool = False
    final_method: Method = Method.AIMNET2_WB97MD3
    solvent: Solvent | None = Solvent.WATER
    max_energy: float = 5

    constraints: list[Constraint] = []


class RdkitConformerSettings(ConformerSettings):
    num_initial_confs: int = 100
    max_mmff_energy: float = 10
    max_mmff_iterations: int = 500


class CrestConformerSettings(ConformerSettings):
    flags: str = "--quick --ewin 10"
    gfn: int | str = "ff"


class Conformer(Base):
    energy: Annotated[float, AfterValidator(round_float(6))]
    weight: Annotated[float | None, AfterValidator(round_optional_float(6))] = None

    uuid: str | None = None


class ConformerWorkflow(MoleculeWorkflow):
    mode: Mode = Mode.RAPID
    settings: ConformerSettings = ConformerSettings()
    conformers: list[Conformer] = []

    def model_post_init(self, __context: Any, /) -> None:
        self.settings = csearch_settings_by_mode(self.mode, self.settings)


def csearch_settings_by_mode(mode: Mode, old_settings: ConformerSettings | None = None) -> ConformerSettings:
    if mode == Mode.MANUAL:
        assert old_settings is not None
        return old_settings

    settings: ConformerSettings

    if mode == Mode.METICULOUS:
        settings = CrestConformerSettings(
            gfn=2,
            flags="--ewin 15 --noreftopo",
            max_energy=10,
            num_confs_considered=500,
            num_confs_taken=150,
        )

    elif mode == Mode.CAREFUL:
        settings = CrestConformerSettings(
            gfn="ff",
            flags="--quick --ewin 10 --noreftopo",
            num_confs_considered=150,
            num_confs_taken=50,
        )

    elif mode == Mode.RAPID or Mode.AUTO:
        settings = RdkitConformerSettings(
            num_initial_confs=300,
            max_mmff_energy=15,
            num_confs_considered=100,
            num_confs_taken=50,
        )

    elif mode == Mode.RECKLESS:
        settings = RdkitConformerSettings(
            num_initial_confs=200,
            max_mmff_energy=10,
            num_confs_considered=50,
            num_confs_taken=20,
            rmsd_cutoff=0.25,
        )

    else:
        raise ValueError(f"invalid mode {mode.value} for conformer settings")

    if old_settings is not None:
        settings.final_method = old_settings.final_method
        settings.solvent = old_settings.solvent
        settings.constraints = old_settings.constraints

    return settings



from pytest import raises

from stjames import Constraint, Method, Mode, OmegaTuning, OptimizationSettings, Settings, SolventSettings, TDDFTSettings


def test_set_mode_auto() -> None:
    Settings()
    assert Settings().mode == Mode.RAPID


def test_opt_settings() -> None:
    settings_rapid = Settings(mode=Mode.RAPID)
    settings_meticulous = Settings(mode=Mode.METICULOUS)

    cons = [Constraint(atoms=[1, 2], constraint_type="bond")]
    settings_careful = Settings(mode=Mode.CAREFUL, opt_settings=OptimizationSettings(constraints=cons))

    rap_opt_set = settings_rapid.opt_settings
    car_opt_set = settings_careful.opt_settings
    met_opt_set = settings_meticulous.opt_settings

    assert not rap_opt_set.constraints
    assert not met_opt_set.constraints
    assert car_opt_set.constraints == cons

    assert rap_opt_set.energy_threshold == 5e-5
    assert rap_opt_set.max_gradient_threshold == 5e-3
    assert rap_opt_set.rms_gradient_threshold == 3.5e-3

    assert car_opt_set.energy_threshold == 1e-6
    assert car_opt_set.max_gradient_threshold == 9e-4
    assert car_opt_set.rms_gradient_threshold == 6e-4

    assert met_opt_set.energy_threshold == 1e-6
    assert met_opt_set.max_gradient_threshold == 3e-5
    assert met_opt_set.rms_gradient_threshold == 2e-5


def test_omega() -> None:
    with raises(ValueError, match="Omega tuning may only be specified for range-separated DFT functionals"):
        Settings(method=Method.B3LYP, omega="koopmans")

    Settings(method=Method.CAMB3LYP, omega="koopmans")
    Settings(method=Method.CAMB3LYP, omega=0.3)


def test_tddft_settings_basic() -> None:
    """Test basic TDDFTSettings construction."""
    settings = Settings(
        method=Method.WB97MD3BJ,
        basis_set="def2-SVP",
        omega="koopmans",
        excited_state_settings=TDDFTSettings(
            tda=False,
            num_excitations=8,
            target_root=2,
        ),
        engine="pyscf",
    )

    assert settings.method == Method.WB97MD3BJ
    assert settings.omega == OmegaTuning.KOOPMANS
    assert isinstance(settings.excited_state_settings, TDDFTSettings)
    assert settings.excited_state_settings.tda is False
    assert settings.excited_state_settings.num_excitations == 8
    assert settings.excited_state_settings.target_root == 2


def test_tddft_settings_custom() -> None:
    """Test TDDFTSettings with custom parameters."""
    settings = Settings(
        method=Method.B3LYP,
        basis_set="sto-3g",
        excited_state_settings=TDDFTSettings(
            tda=True,
            num_excitations=10,
            target_root=5,
        ),
        engine="gpu4pyscf",
    )

    assert settings.method == Method.B3LYP
    assert settings.omega is None
    assert isinstance(settings.excited_state_settings, TDDFTSettings)
    assert settings.excited_state_settings.tda is True
    assert settings.excited_state_settings.num_excitations == 10
    assert settings.excited_state_settings.target_root == 5


def test_settings_roundtrip() -> None:
    settings = Settings(
        method=Method.B3LYP,
        basis_set="sto-3g",
        excited_state_settings=TDDFTSettings(num_excitations=5),
        engine="pyscf",
    )

    data = settings.model_dump(mode="json")
    assert isinstance(data, dict)
    assert isinstance(data["excited_state_settings"], dict)

    stjames_settings = Settings.model_validate(data)

    assert stjames_settings.method == settings.method
    assert stjames_settings.basis_set == settings.basis_set
    assert stjames_settings.engine == settings.engine
    assert isinstance(stjames_settings.excited_state_settings, TDDFTSettings)
    assert stjames_settings.excited_state_settings == settings.excited_state_settings


def test_gpu4pyscf_solvent() -> None:
    with raises(ValueError, match="GPU4PySCF does not support the COSMO solvent model"):
        Settings(
            method=Method.B3LYP,
            basis_set="sto-3g",
            solvent_settings=SolventSettings(model="COSMO", solvent="hexane"),
            engine="gpu4pyscf",
        )



"""Solubility prediction workflow."""

from typing import Annotated, Self

from pydantic import AfterValidator, BaseModel, model_validator

from ..base import LowercaseStrEnum
from ..types import round_list
from .workflow import SMILESWorkflow


class SolubilityMethod(LowercaseStrEnum):
    FASTSOLV = "fastsolv"
    KINGFISHER = "kingfisher"
    ESOL = "esol"


class SolubilityResult(BaseModel):
    """
    Solubility result.

    :param solubilities: solubilities in log(S)/L
    :param uncertainties: uncertainties in the solubility in log(S)/L
    """

    solubilities: Annotated[list[float], AfterValidator(round_list(6))]
    uncertainties: Annotated[list[float | None], AfterValidator(round_list(6))]

    @model_validator(mode="after")
    def check_size(self) -> Self:
        """Confirm that the shapes are the same"""
        if len(self.solubilities) != len(self.uncertainties):
            raise ValueError("Solubilities and uncertainties must have the same length.")

        return self


class SolubilityWorkflow(SMILESWorkflow):
    """
    Solubility prediction workflow.

    Inputs:
    :param initial_smiles: SMILES string of the molecule
    :param solubility_method: model used for solubility prediction
    :param solvents: list of solvent SMILES strings
    :param temperatures: temperatures, in K

    Results:
    :param solubilities: {solvent: SolubilityResult}
    """

    initial_smiles: str
    solubility_method: SolubilityMethod = SolubilityMethod.FASTSOLV
    solvents: list[str] = ["O"]
    temperatures: list[float] = [298.15]

    solubilities: dict[str, SolubilityResult] = {}

    @model_validator(mode="after")
    def check_size(self) -> Self:
        """Check that the sizes of the lists are consistent."""
        for solvent, result in self.solubilities.items():
            if len(result.solubilities) != len(self.temperatures):
                raise ValueError(f"Solubilities for {solvent} must have the same length as temperatures.")

            if solvent not in self.solvents:
                raise ValueError(f"Solvent {solvent} not in initial list of solvents.")

        return self

    @model_validator(mode="after")
    def check_solvent_temperature(self) -> Self:
        """Check that models with limited domain of applicability are predicting within correct domain."""
        match self.solubility_method:
            case SolubilityMethod.KINGFISHER | SolubilityMethod.ESOL:
                if (len(self.solvents) > 1) or (self.solvents[0] != "O"):
                    raise ValueError(f"Method `{self.solubility_method}` can only predict aqueous solubility, so `solvents` must be [`O`] only.")
                if (len(self.temperatures) > 1) or (abs(self.temperatures[0] - 298.15) > 0.1):
                    raise ValueError(
                        f"Method `{self.solubility_method}` can only predict solubility at room temperature, so `temperatures` must be [298.15] only."
                    )
            case _:
                pass

        return self



from typing import Annotated

from pydantic import AfterValidator, NonNegativeInt

from ..base import Base, round_float
from ..types import Vector3D
from .workflow import ProteinStructureWorkflow


class Pocket(Base):
    """
    Represents a pocket.

    :param sphere_centers: centers of detected spheres
    :param sphere_radii: radii of detected spheres
    :param volume: volume, in Å³
    :param score: druggability/quality score; larger scores are better
    :param pocket_center: center of bounding box
    :param pocket_sides: side lengths of bounding box
    :param residue_numbers: indices of residues on pocket
    """

    sphere_centers: list[Vector3D]
    sphere_radii: list[float]

    volume: Annotated[float, AfterValidator(round_float(3))]
    score: Annotated[float, AfterValidator(round_float(3))]

    pocket_center: Vector3D
    pocket_sides: Vector3D

    residue_numbers: list[NonNegativeInt]


class PocketDetectionWorkflow(ProteinStructureWorkflow):
    """
    Uses Pocketeer to detect potential binding sites on a protein.

    Inherited:
    :param protein: protein

    New:
    :param merge_distance: distance for merging pocket spheres, in Å

    Results:
    :param pockets: located pockets
    """

    merge_distance: Annotated[float, AfterValidator(round_float(3))] = 1.75
    pockets: list[Pocket] = []



"""Double ended transition-state-search workflow."""

from typing import Annotated, Self

from pydantic import AfterValidator, PositiveFloat, model_validator

from ..molecule import Molecule
from ..optimization.freezing_string_method import FSMSettings
from ..settings import Settings
from ..types import UUID, round_list
from .workflow import Workflow


class DoubleEndedTSSearchWorkflow(Workflow):
    """
    Settings for running a double-ended-transition-state-search Workflow.

    # Inputs
    :param reactant: reactant Molecule
    :param product: product Molecule
    :param calculation_settings: Settings for the calculations
    :param search_settings: Settings for the search
    :param optimize_inputs: Whether to optimize the input reactant and product
    :param optimize_ts: Whether to optimize the guess transition state

    # Results
    :param forward_string_distances: distances along the string
    :param backward_string_distances: distances along the backward string (starting from the end of the forward string)
    :param forward_calculation_uuids: UUIDs of the calculations for the forward string nodes
    :param backward_calculation_uuids: UUIDs of the calculations for the backward string nodes
    :param ts_guess_calculation_uuid: UUID of the calculation for the guess (or optimized) transition state
    """

    reactant: Molecule
    product: Molecule

    calculation_settings: Settings
    search_settings: FSMSettings
    optimize_inputs: bool = False
    optimize_ts: bool = True

    # Results
    forward_string_distances: Annotated[list[PositiveFloat], AfterValidator(round_list(5))] = []
    backward_string_distances: Annotated[list[PositiveFloat], AfterValidator(round_list(5))] = []

    forward_calculation_uuids: list[UUID | None] = []
    backward_calculation_uuids: list[UUID | None] = []

    ts_guess_calculation_uuid: UUID | None = None

    @model_validator(mode="after")
    def validate_input_molecules(self) -> Self:
        """Validate that reactant and product are compatible."""

        if self.reactant.charge != self.product.charge:
            raise ValueError("Reactant and product must have the same charge.")
        if self.reactant.multiplicity != self.product.multiplicity:
            raise ValueError("Reactant and product must have the same multiplicity.")

        if len(self.reactant) != len(self.product):
            raise ValueError("Reactant and product must have the same number of atoms.")

        for r_atom, p_atom in zip(self.reactant.atoms, self.product.atoms, strict=True):
            if r_atom.atomic_number != p_atom.atomic_number:
                raise ValueError("Reactant and product must have the same atom ordering.")

        return self

    @property
    def path_uuids(self) -> list[UUID | None]:
        """Return the path from reactant to product."""
        return self.forward_calculation_uuids + self.backward_calculation_uuids

    @property
    def distances(self) -> list[float]:
        """Return the path from reactant to product."""
        return self.forward_string_distances + self.backward_string_distances[::-1]



import pydantic

from .base import Base


class ThermochemistrySettings(Base):
    """
    Thermochemistry calculation settings.

    :param cutoff_frequency: Cramer/Truhlar quasi-harmonic cutoff, in cm^-1
    :param temperature: temperature for thermochemistry, in K
    :param scaling_factor: frequency scaling factor
    :param concentration: concentration, in M (defaults to 1 atm)
    """

    # Cramer/Truhlar cutoff freq (cm^-1)
    cutoff_frequency: pydantic.NonNegativeFloat = 100

    # temp (K)
    temperature: pydantic.NonNegativeFloat = 298

    # scaling factor, for frequencies
    scaling_factor: pydantic.NonNegativeFloat = 1.0

    # conc (M), defaults to 1 atm
    # (cf. https://github.com/patonlab/GoodVibes/blob/master/goodvibes/GoodVibes.py#L403)
    concentration: pydantic.NonNegativeFloat = 0.0408740470708



from typing import Annotated, Literal

from pydantic import AfterValidator, PositiveInt

from ..base import Base, LowercaseStrEnum, round_float
from ..basis_set import BasisSet
from .workflow import MoleculeWorkflow


class EnergyDecompositionMethod(LowercaseStrEnum):
    SAPT0 = "sapt0"


class EnergyDecompositionSettings(Base):
    """
    How energy-decomposition calculations should be performed.

    :param method: which energy decomposition method
    :param basis_set: which basis set to employ
    """

    method: EnergyDecompositionMethod = EnergyDecompositionMethod.SAPT0
    basis_set: BasisSet = BasisSet(name="jun-cc-pVDZ")

    def __str__(self) -> str:
        """
        >>> str(EnergyDecompositionSettings())
        'SAPT0(jun-cc-pVDZ)'
        """
        return f"{self.method.value.upper()}({self.basis_set.name})"


class SAPT0Result(Base):
    """
    Stores the result of a SAPT0 calculation.

    :param total_interaction_energy: total interaction energy, in kcal/mol
    :param electrostatic_interaction_energy: electrostatic interaction energy, in kcal/mol
    :param exchange_interaction_energy: exchange interaction energy, in kcal/mol
    :param dispersion_interaction_energy: dispersion interaction energy, in kcal/mol
    :param induction_interaction_energy: induction interaction energy, in kcal/mol
    :param result_type: to disambiguate from possible future alternate returns
    """

    total_interaction_energy: Annotated[float, AfterValidator(round_float(3))]
    electrostatic_interaction_energy: Annotated[float, AfterValidator(round_float(3))]
    exchange_interaction_energy: Annotated[float, AfterValidator(round_float(3))]
    dispersion_interaction_energy: Annotated[float, AfterValidator(round_float(3))]
    induction_interaction_energy: Annotated[float, AfterValidator(round_float(3))]

    result_type: Literal["sapt0"] = "sapt0"


class InteractionEnergyDecompositionWorkflow(MoleculeWorkflow):
    """
    Performs an interaction-energy-decomposition calculation.

    Inherited:
    :param initial_molecule: Molecule in question

    New:
    :param fragment1_indices: which atoms go to fragment #1, fragment #2 takes the rest (1-indexed)
    :param energy_decomposition_settings: settings for SAPT calculations

    Results:
    :param energy_decomposition_result: results from the energy decomposition
    """

    fragment1_indices: list[PositiveInt]

    energy_decomposition_settings: EnergyDecompositionSettings = EnergyDecompositionSettings()

    energy_decomposition_result: SAPT0Result | None = None



from typing import Sequence

from pydantic import PositiveFloat, PositiveInt

from .base import Base
from .constraint import Constraint


class OptimizationSettings(Base):
    """
    Geometry optimization settings.

    :param max_steps: maximum number of optimization steps
    :param transition_state: perform transition state optimization
    :param recalc_hess_every: recalculate Hessian every n steps (0 = never)
    :param max_gradient_threshold: convergence threshold for max gradient, in Hartree/Å
    :param rms_gradient_threshold: convergence threshold for RMS gradient, in Hartree/Å
    :param energy_threshold: convergence threshold for energy change, in Hartree
    :param optimize_cell: optimize unit cell (periodic systems only)
    :param constraints: geometric constraints to apply
    :param save_intermediate_steps: save intermediate geometries
    """

    max_steps: PositiveInt = 250
    transition_state: bool = False
    recalc_hess_every: int = 0

    max_gradient_threshold: PositiveFloat = 7e-4
    rms_gradient_threshold: PositiveFloat = 6e-4
    energy_threshold: PositiveFloat = 1e-6

    optimize_cell: bool = False

    constraints: Sequence[Constraint] = ()

    save_intermediate_steps: bool = True



"""Base classes for workflows."""

from typing import Any

from pydantic import field_validator

from ..base import Base
from ..dna import DNASequence
from ..message import Message
from ..mode import Mode
from ..molecule import Molecule
from ..pdb import PDB
from ..protein import ProteinSequence
from ..rna import RNASequence
from ..types import UUID


class Workflow(Base):
    """
    Base class for Workflows.

    :param messages: messages to display
    """

    messages: list[Message] = []

    def __str__(self) -> str:
        return repr(self)


class FASTAWorkflow(Workflow):
    """
    Base class for Workflows that operate on biological sequences and SMILES.

    :param initial_protein_sequences: protein sequences to evaluate, either plain sequence strings or ProteinSequence objects with metadata
    :param initial_dna_sequences: DNA sequences to evaluate, either plain sequence strings or DNASequence objects with metadata
    :param initial_rna_sequences: RNA sequences to evaluate, either plain sequence strings or RNASequence objects with metadata
    :param initial_smiles_list: SMILES strings of interest
    :param ligand_binding_affinity_index: optional index selecting which ligand affinity to evaluate
    :raises ValueError: if none of the sequence lists are provided
    """

    initial_protein_sequences: list[ProteinSequence] | list[str] = []
    initial_dna_sequences: list[DNASequence] = []
    initial_rna_sequences: list[RNASequence] = []
    initial_smiles_list: list[str] = []
    ligand_binding_affinity_index: int | None = None

    def model_post_init(self, __context: Any, /) -> None:
        if not (self.initial_protein_sequences or self.initial_dna_sequences or self.initial_rna_sequences):
            raise ValueError(
                "Provide at least one of `initial_protein_sequences`, `initial_dna_sequences`, or `initial_rna_sequences`.",
            )


class SMILESWorkflow(Workflow):
    """
    Base class for Workflows that operate on a single SMILES string.

    :param initial_smiles: SMILES string of interest
    """

    initial_smiles: str

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.initial_smiles}>"


class BatchSMILESWorkflow(Workflow):
    """
    Base class for Workflows that operate on a list of SMILES strings.

    :param initial_smiles_list: SMILES strings of interest
    """

    initial_smiles_list: list[str]


class MoleculeWorkflow(Workflow):
    """
    Base class for Workflows that operate on a single molecule.

    :param initial_molecule: Molecule of interest
    :param mode: Mode to use
    """

    initial_molecule: Molecule
    mode: Mode = Mode.AUTO

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.mode.name}>"

    @field_validator("mode")
    @classmethod
    def set_mode_auto(cls, mode: Mode) -> Mode:
        """Set the mode to RAPID if AUTO is selected."""
        if mode == Mode.AUTO:
            return Mode.RAPID

        return mode


class ProteinSequenceWorkflow(Workflow):
    """
    Base class for Workflows that operate on protein sequences.

    :param initial_protein_sequences: protein sequences to evaluate, either plain sequence strings or ProteinSequence objects with metadata
    """

    initial_protein_sequences: list[ProteinSequence] | list[str] = []


class ProteinStructureWorkflow(Workflow):
    """
    Base class for Workflows that operate on protein sequences.

    :param protein: PDB structure of protein, or UUID of PDB record in Rowan
    """

    protein: PDB | UUID


class DBCalculation(Base):
    """Encodes a calculation that's in the database. This isn't terribly useful by itself."""

    uuid: UUID



from typing import Annotated, Any

from pydantic import AfterValidator, PositiveFloat, PositiveInt, model_validator

from ..base import Base, round_float
from ..types import UUID, ProteinMDTrajectory, round_list
from .workflow import ProteinStructureWorkflow, SMILESWorkflow


class BindingPoseContact(Base):
    """
    A single protein–ligand contact from an MD trajectory.

    :param protein_atom_index: index of protein atom
    :param ligand_atom_index: index of ligand atom
    :occupancy: the probability of seeing this interaction in a frame, between 0 and 1
    """

    protein_atom_index: int
    ligand_atom_index: int
    occupancy: Annotated[float, AfterValidator(round_float(3))]


class BindingPoseTrajectory(ProteinMDTrajectory):
    """
    Represents a single trajectory looking at a binding pose.

    Inherited:
    :param uuid: UUID of trajectory

    New:
    :param ligand_rmsd: RMSD of ligand vs. starting pose (aligning protein)
    :param contacts: conserved binding-pose contacts
    """

    ligand_rmsd: Annotated[list[float], AfterValidator(round_list(3))] = []
    contacts: list[BindingPoseContact] = []


class ProteinMDSettingsMixin(Base):
    """
    Mix-in for various settings used in running protein MD.

    :param equilibration_time_ns: how long to equilibrate trajectories for, in ns
    :param simulation_time_ns: how long to run trajectories for, in ns
    :param temperature: temperature, in K
    :param pressure_atm: pressure, in atm
    :param langevin_timescale_ps: timescale for the Langevin integrator, in ps⁻¹
    :param timestep_fs: timestep, in femtoseconds
    :param constrain_hydrogens: whether or not to use SHAKE to freeze bonds to hydrogen
    :param nonbonded_cutoff: nonbonded cutoff for particle-mesh Ewald, in Å
    :param ionic_strength_M: ionic strength of the solution, in M (molar)
    :param water_buffer: amount of water to add around the protein, in Å
    :param num_clusters: the number of clusters to use for trajectory post-processing
    """

    equilibration_time_ns: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 1
    simulation_time_ns: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 10

    temperature: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 300
    pressure_atm: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 1.0
    langevin_timescale_ps: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 1.0

    timestep_fs: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 2
    constrain_hydrogens: bool = True
    nonbonded_cutoff: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 8.0

    ionic_strength_M: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 0.10
    water_buffer: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 6.0

    num_clusters: PositiveInt = 10


class ProteinMolecularDynamicsWorkflow(ProteinMDSettingsMixin, ProteinStructureWorkflow):
    """
    Protein molecular dynamics workflow.

    Inherited:
    :param protein: PDB or UUID of the (holo) protein.
    :param equilibration_time_ns: how long to equilibrate trajectories for, in ns
    :param simulation_time_ns: how long to run trajectories for, in ns
    :param temperature: temperature, in K
    :param pressure_atm: pressure, in atm
    :param langevin_timescale_ps: timescale for Langevin integrator, in ps⁻¹
    :param timestep_fs: timestep, in femtoseconds
    :param constrain_hydrogens: whether to use SHAKE to freeze bonds to hydrogen
    :param nonbonded_cutoff: nonbonded cutoff for particle-mesh Ewald, in Å
    :param protein_restraint_cutoff: cutoff past which alpha-carbons constrained, in Å
    :param protein_restraint_constant: force constant for backbone restraints, in kcal/mol/Å²
    :param ionic_strength_M: ionic strength of solution, in M (molar)
    :param water_buffer: amount of water to add around protein, in Å

    New:
    :param num_trajectories: number of trajectories to run
    :param save_solvent: whether solvent should be saved

    Results:
    :param minimized_protein_uuid: UUID of final system PDB
    :param bonds: which atoms are bonded to which other atoms
    :param trajectories: UUID for each trajectory
    """

    num_trajectories: PositiveInt = 1
    save_solvent: bool = False

    minimized_protein_uuid: UUID | None = None
    bonds: list[tuple[int, int]] = []
    trajectories: list[ProteinMDTrajectory] = []


class PoseAnalysisMolecularDynamicsWorkflow(ProteinMDSettingsMixin, ProteinStructureWorkflow, SMILESWorkflow):
    """
    Pose-analysis molecular dynamics workflow.

    Inherited:
    :param initial_smiles: ligand's SMILES
    :param protein: PDB or UUID of the (holo) protein.
    :param equilibration_time_ns: how long to equilibrate trajectories for, in ns
    :param simulation_time_ns: how long to run trajectories for, in ns
    :param temperature: temperature, in K
    :param pressure_atm: pressure, in atm
    :param langevin_timescale_ps: timescale for the Langevin integrator, in ps⁻¹
    :param timestep_fs: timestep, in femtoseconds
    :param constrain_hydrogens: whether or not to use SHAKE to freeze bonds to hydrogen
    :param nonbonded_cutoff: nonbonded cutoff for particle-mesh Ewald, in Å
    :param ionic_strength_M: ionic strength of the solution, in M (molar)
    :param water_buffer: amount of water to add around the protein, in Å

    New:
    :param protein_uuid: UUID of the (holo) protein. DEPRECATED.
    :param ligand_residue_name: ligand's residue name
    :param num_trajectories: number of trajectories to run
    :param save_solvent: whether any solvent should be saved
    :param max_num_solvent: saves this many solvent molecules, or all solvent molecules if None
    :param protein_restraint_cutoff: cutoff past which alpha-carbons will be constrained, in Å
    :param protein_restraint_constant: force constant for backbone restraints, in kcal/mol/Å²

    Results:
    :param minimized_protein_uuid: UUID of final system PDB
    :param bonds: which atoms are bonded to which other atoms
    :param trajectories: for each replicate, a UUID and the corresponding analysis results
    """

    protein_uuid: UUID | None = None
    ligand_residue_name: str = "LIG"

    num_trajectories: PositiveInt = 1
    save_solvent: bool = False
    num_solvent_to_save: PositiveInt | None = None

    protein_restraint_cutoff: Annotated[PositiveFloat, AfterValidator(round_float(3))] | None = None
    protein_restraint_constant: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 100

    minimized_protein_uuid: UUID | None = None
    bonds: list[tuple[int, int]] = []
    trajectories: list[BindingPoseTrajectory] = []

    @model_validator(mode="before")
    def harmonize_protein_uuid(cls, data: Any) -> Any:  # noqa: N805
        """
        Syncs data between "protein_uuid" and "protein" field.
        """
        if "protein_uuid" in data:
            data["protein"] = data["protein_uuid"]

        return data



"""High-throughput docking workflow."""

from typing import Annotated, Any

from pydantic import AfterValidator, ConfigDict, field_validator, model_validator

from ..pdb import PDB
from ..types import UUID, Vector3D, round_list
from .docking import VinaSettings
from .workflow import BatchSMILESWorkflow, ProteinStructureWorkflow


class BatchDockingWorkflow(BatchSMILESWorkflow, ProteinStructureWorkflow):
    """
    Batch docking workflow.

    Inherited:
    :param initial_smiles_list: list of SMILES
    :param protein: PDB of the protein, or the UUID of the protein.

    New:
    :param target: PDB of the protein, or the UUID of the protein. DEPRECATED
    :param pocket: center (x, y, z) and size (x, y, z) of the pocket
    :param docking_settings: how to run each docking calculation

    Results:
    :param best_scores: best score for each SMILES string
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    target: PDB | UUID
    pocket: tuple[Vector3D, Vector3D]

    docking_settings: VinaSettings = VinaSettings()
    best_scores: Annotated[list[float | None], AfterValidator(round_list(3))] = []

    @model_validator(mode="before")
    def harmonize_target_and_protein(cls, data: Any) -> Any:  # noqa: N805
        """
        Syncs data between "target" and "protein" field.
        """
        protein = data.get("protein")
        target = data.get("target")

        if target and not protein:
            data["protein"] = target
        elif not target and protein:
            data["target"] = protein

        return data

    @field_validator("pocket", mode="after")
    def validate_pocket(cls, pocket: tuple[Vector3D, Vector3D]) -> tuple[Vector3D, Vector3D]:
        _center, size = pocket
        if any(q <= 0 for q in size):
            raise ValueError(f"Pocket size must be positive, got: {size}")
        return pocket



import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Iterable, Self, Sequence, TypeAlias, TypedDict, TypeVar

import numpy as np
import pydantic
from pydantic import AfterValidator, NonNegativeInt, PositiveInt, ValidationError

from .atom import Atom
from .base import Base, round_float, round_optional_float
from .data import SYMBOL_ELEMENT
from .periodic_cell import PeriodicCell
from .types import (
    FloatPerAtom,
    Matrix3x3,
    Vector3D,
    Vector3DPerAtom,
    round_list,
    round_optional_float_per_atom,
    round_optional_matrix3x3,
    round_optional_vector3d,
    round_optional_vector3d_per_atom,
    round_vector3d_per_atom,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rdkit import Chem

    RdkitMol: TypeAlias = Chem.rdchem.Mol | Chem.rdchem.RWMol
else:
    RdkitMol = Any


class MoleculeReadError(RuntimeError):
    """Error raised when molecule parsing fails."""


class VibrationalMode(Base):
    """
    Vibrational mode from frequency analysis.

    :param frequency: vibrational frequency, in cm⁻¹
    :param reduced_mass: reduced mass, in amu
    :param force_constant: force constant, in mDyne/Å
    :param displacements: atomic displacement vectors, in Å
    :param ir_intensity: infrared intensity, in km/mol
    """

    frequency: Annotated[float, AfterValidator(round_float(3))]
    reduced_mass: Annotated[float, AfterValidator(round_float(3))]
    force_constant: Annotated[float, AfterValidator(round_float(3))]
    displacements: Annotated[Vector3DPerAtom, AfterValidator(round_vector3d_per_atom(6))]
    ir_intensity: Annotated[float | None, AfterValidator(round_optional_float(3))] = None


class Molecule(Base):
    """
    A molecular structure with associated properties.

    :param charge: total charge of the Molecule
    :param multiplicity: spin multiplicity of the Molecule
    :param atoms: Atom objects representing the atoms in the Molecule
    :param cell: PeriodicCell for periodic boundary conditions
    :param energy: electronic energy, in Hartree
    :param scf_iterations: number of SCF iterations
    :param scf_completed: whether the SCF converged
    :param elapsed: time taken for the calculation, in seconds
    :param homo_lumo_gap: energy of the HOMO-LUMO gap, in eV
    :param gradient: energy gradient with respect to position, in Hartree/Å
    :param stress: stress matrix
    :param velocities: velocities of the atoms, in Å/fs
    :param mulliken_charges: Mulliken charges
    :param mulliken_spin_densities: Mulliken spin densities
    :param dipole: dipole moment, in Debye
    :param vibrational_modes: vibrations
    # https://docs.rowansci.com/science/quantum-chemistry/frequencies-and-thermochemistry
    :param zero_point_energy: zero-point energy, in Hartree
    :param thermal_energy_corr: ZPE + non-zero-temperature effects, in Hartree
    :param thermal_enthalpy_corr: thermal_energy_corr + pV, in Hartree
    :param thermal_free_energy_corr: thermal_enthalpy_corr - 298.15 x S, in Hartree
    :param excitation_energies: electronic-excitation energies, in Hartree
    :param oscillator_strengths: electronic-transition strengths, dimensionless
    :param smiles: SMILES corresponding to the Molecule
    :param calculation_index: index in a calculation output
    """

    charge: int
    multiplicity: PositiveInt
    atoms: list[Atom]

    cell: PeriodicCell | None = None

    energy: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    scf_iterations: NonNegativeInt | None = None
    scf_completed: bool | None = None
    elapsed: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    homo_lumo_gap: Annotated[float | None, AfterValidator(round_optional_float(6))] = None

    gradient: Annotated[Vector3DPerAtom | None, AfterValidator(round_optional_vector3d_per_atom(6))] = None
    stress: Annotated[Matrix3x3 | None, AfterValidator(round_optional_matrix3x3(6))] = None  # Hartree/Å

    velocities: Annotated[Vector3DPerAtom | None, AfterValidator(round_optional_vector3d_per_atom(6))] = None

    mulliken_charges: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    mulliken_spin_densities: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    dipole: Annotated[Vector3D | None, AfterValidator(round_optional_vector3d(6))] = None

    vibrational_modes: list[VibrationalMode] | None = None

    zero_point_energy: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    thermal_energy_corr: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    thermal_enthalpy_corr: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    thermal_free_energy_corr: Annotated[float | None, AfterValidator(round_optional_float(6))] = None

    excitation_energies: Annotated[list[float] | None, AfterValidator(round_list(3))] = None
    oscillator_strengths: Annotated[list[float] | None, AfterValidator(round_list(3))] = None

    smiles: str | None = None
    calculation_index: int | None = None

    def __len__(self) -> int:
        return len(self.atoms)

    def distance(self, i: PositiveInt, j: PositiveInt) -> float:
        r"""
        Calculate the distance between atoms, in Å.

        >>> mol = Molecule.from_xyz("H 0 1 0\nH 0 0 1")
        >>> mol.distance(1, 2)
        1.4142135623730951
        """
        return sum((q2 - q1) ** 2 for q1, q2 in zip(self.atoms[i - 1].position, self.atoms[j - 1].position, strict=True)) ** 0.5  # type: ignore [no-any-return,unused-ignore]

    def angle(self, i: PositiveInt, j: PositiveInt, k: PositiveInt, degrees: bool = True) -> float:
        r"""
        Calculate the angle between three atoms.

        >>> Molecule.from_xyz("H 0 0 0\nO 0 0 1\nH 0 1 1").angle(1, 2, 3)
        90.0
        """

        return angle(self.coordinates[i - 1], self.coordinates[j - 1], self.coordinates[k - 1], degrees=degrees)

    def dihedral(self, i: int, j: int, k: int, l: int, degrees: bool = True, positive_domain: bool = True) -> float:
        r"""
        Calculate the dihedral angle between four atoms.

        >>> Molecule.from_xyz("H 0 0 0\nO 0 0 1\nO 0 1 1\nH 1 1 1").dihedral(1, 2, 3, 4)
        270.0
        """
        return dihedral(
            self.coordinates[i - 1],
            self.coordinates[j - 1],
            self.coordinates[k - 1],
            self.coordinates[l - 1],
            degrees=degrees,
            positive_domain=positive_domain,
        )

    @property
    def coordinates(self) -> Vector3DPerAtom:
        """Cartesian coordinates of all atoms, in Å."""
        return [a.position for a in self.atoms]

    def translated(self, vector: Vector3D) -> Self:
        r"""
        Translate the molecule by a vector, in Å.

        >>> mol = Molecule.from_xyz("H 0 0 0\nH 0 0 1")
        >>> print(mol.translated((1, 0, 0)).to_xyz())
        2
        charge: 0; multiplicity: 1;
        H     1.0000000000    0.0000000000    0.0000000000
        H     1.0000000000    0.0000000000    1.0000000000
        """

        def translated(position: Vector3D) -> Vector3D:
            return tuple(q + v for q, v in zip(position, vector, strict=True))  # type: ignore [return-value]

        atoms = [atom.model_copy(update={"position": translated(atom.position)}) for atom in self.atoms]

        return self.model_copy(update={"atoms": atoms})

    @property
    def atomic_numbers(self) -> list[NonNegativeInt]:
        """Atomic numbers of all atoms."""
        return [a.atomic_number for a in self.atoms]

    @property
    def sum_energy_zpe(self) -> float | None:
        """Electronic energy plus zero-point energy, in Hartree."""
        if (self.energy is None) or (self.zero_point_energy is None):
            return None
        return self.energy + self.zero_point_energy

    @property
    def sum_energy_thermal_corr(self) -> float | None:
        """Electronic energy plus thermal energy correction, in Hartree."""
        if (self.energy is None) or (self.thermal_energy_corr is None):
            return None
        return self.energy + self.thermal_energy_corr

    @property
    def sum_energy_enthalpy(self) -> float | None:
        """Electronic energy plus enthalpy correction, in Hartree."""
        if (self.energy is None) or (self.thermal_enthalpy_corr is None):
            return None
        return self.energy + self.thermal_enthalpy_corr

    @property
    def enthalpy(self) -> float | None:
        """Alias for sum_energy_enthalpy, in Hartree."""
        return self.sum_energy_enthalpy

    @property
    def sum_energy_free_energy(self) -> float | None:
        """Electronic energy plus Gibbs free energy correction, in Hartree."""
        if (self.energy is None) or (self.thermal_free_energy_corr is None):
            return None
        return self.energy + self.thermal_free_energy_corr

    @property
    def gibbs_free_energy(self) -> float | None:
        """Alias for sum_energy_free_energy, in Hartree."""
        return self.sum_energy_free_energy

    @pydantic.model_validator(mode="after")
    def check_electron_sanity(self) -> Self:
        """Check that the charge and multiplicity combination is possible."""
        num_electrons = sum(self.atomic_numbers) - self.charge
        num_unpaired_electrons = self.multiplicity - 1
        if (num_electrons - num_unpaired_electrons) % 2 != 0:
            raise ValueError(
                f"The combination of {num_electrons} electrons, charge {self.charge}, and multiplicity {self.multiplicity} is impossible. "
                "Double-check the charge and multiplicity values given and verify that they are correct."
            )

        return self

    @classmethod
    def from_file(cls: type[Self], filename: Path | str, format: str | None = None, charge: int | None = None, multiplicity: PositiveInt | None = None) -> Self:
        r"""
        Read a molecule from a file.

        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile("w+", suffix=".xyz") as f:
        ...    _ = f.write("2\nComment\nH 0 0 0\nF 0 0 1")
        ...    _ = f.seek(0)
        ...    mol = Molecule.from_file(f.name)
        >>> print(mol.to_xyz())
        2
        charge: 0; multiplicity: 1;
        H     0.0000000000    0.0000000000    0.0000000000
        F     0.0000000000    0.0000000000    1.0000000000
        """
        filename = Path(filename)
        if not format:
            format = filename.suffix[1:]

        with open(filename) as f:
            match format:
                case "xyz":
                    return cls.from_xyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case "extxyz":
                    return cls.from_extxyz_lines(f.readlines(), charge=charge, multiplicity=multiplicity)
                case _:
                    raise ValueError(f"Unsupported {format=}")

    @classmethod
    def from_xyz(cls: type[Self], xyz: str, charge: int | None = None, multiplicity: PositiveInt | None = None) -> Self:
        r"""
        Generate a Molecule from an XYZ string.

        Note: only supports single molecule inputs.

        >>> len(Molecule.from_xyz("2\nComment\nH 0 0 0\nH 0 0 1"))
        2
        """
        return cls.from_xyz_lines(xyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_xyz_lines(cls: type[Self], lines: Iterable[str], charge: int | None = None, multiplicity: PositiveInt | None = None) -> Self:
        r"""
        Read a molecule from a xyz lines.

        >>> mol = Molecule.from_xyz_lines(["2", "charge: 0; multiplicity: 1; cell: [[1, 2e1, 3], [4, 5, 6], [7, 8, 9.1]]", "H 0 0 0", "F 0 0 1"])
        >>> print(mol.to_xyz())
        2
        charge: 0; multiplicity: 1; cell: ((1.0, 20.0, 3.0), (4.0, 5.0, 6.0), (7.0, 8.0, 9.1)); is_periodic: (True, True, True);
        H     0.0000000000    0.0000000000    0.0000000000
        F     0.0000000000    0.0000000000    1.0000000000
        >>> mol = Molecule.from_xyz_lines(["2", "energy: abc", "H 0 0 0", "F 0 0 1"])
        >>> print(mol.to_xyz())
        2
        charge: 0; multiplicity: 1;
        H     0.0000000000    0.0000000000    0.0000000000
        F     0.0000000000    0.0000000000    1.0000000000
        """
        lines = list(lines)
        data: dict[str, Any] = {}
        if len(lines[0].split()) == 1:
            natoms, comment, *lines = lines
            if (not natoms.strip().isdigit()) or (int(natoms) != len(lines)):
                raise MoleculeReadError(f"First line of XYZ file should be the number of atoms ({len(lines)}), got: {natoms}")

            data = cls._parse_comment_line(comment)

        charge = charge if charge is not None else data.get("charge", 0)
        multiplicity = multiplicity or data.get("multiplicity", 1)
        data |= {"charge": charge, "multiplicity": multiplicity}

        try:
            return cls(atoms=[Atom.from_xyz(line) for line in lines], **data)
        except (ValueError, ValidationError):
            pass

        try:
            return cls(atoms=[Atom.from_xyz(line) for line in lines], charge=charge, multiplicity=multiplicity)
        except (ValueError, ValidationError) as e:
            raise MoleculeReadError("Error reading molecule from xyz") from e

    @classmethod
    def _parse_comment_line(cls, comment: str) -> dict[str, Any]:
        """
        Parse the comment line of an XYZ file.

        :param comment: comment line from an XYZ file

        >>> Molecule._parse_comment_line("charge: -1; multiplicity: 2; cell: [[1,2,3],[4,5,6],[7,8,9]]; is_periodic: (True, False, True); energy: -75.0")
        {'charge': '-1', 'multiplicity': '2', 'energy': '-75.0', 'cell':\
        PeriodicCell(lattice_vectors=((1.0, 2.0, 3.0), (4.0, 5.0, 6.0), (7.0, 8.0, 9.0)), is_periodic=(True, False, True))}
        >>> Molecule._parse_comment_line(" energy: -0.320207535977 gnorm: 0.071552110436 xtb: 6.6.1 (8d0f1dd)")  # Unfortunate
        {'energy': '-0.320207535977 gnorm: 0.071552110436 xtb: 6.6.1 (8d0f1dd)'}
        """
        data: dict[str, Any] = {}
        for kv in comment.strip(";").split(";"):
            try:
                key, value = kv.split(":", 1)
                data[key.strip()] = value.strip()
            except ValueError:
                logger.error(f"Error parsing key/value: {kv}")
                continue

        if cell := data.pop("cell", None):
            try:
                data["cell"] = PeriodicCell.from_string(cell)
                if is_periodic := data.pop("is_periodic", None):
                    x, y, z = is_periodic.strip(" ([)]").split(",")
                    true_values = {"true", "1", "yes"}
                    data["cell"].is_periodic = tuple((v.strip().lower() in true_values) for v in (x, y, z))
            except ValueError as e:
                logger.error(f"Error parsing XYZ cell: {e}")

        return data

    def to_xyz(self, comment: str | None = None, out_file: Path | str | None = None) -> str:
        r"""
        Generate an XYZ string.

        :param comment: optional comment line (defaults to standard Rowan XYZ comment line format)
        :param out_file: optional output file path to write the XYZ string to
        :return: XYZ string

        >>> mol = Molecule.from_xyz("2\nenergy: 1.0; smiles: HF;\nH 0 1 2\nF 1 2 3")
        >>> print(mol.to_xyz())
        2
        charge: 0; multiplicity: 1; energy: 1.0; smiles: HF;
        H     0.0000000000    1.0000000000    2.0000000000
        F     1.0000000000    2.0000000000    3.0000000000
        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as directory:
        ...     file = Path(directory) / "mol.xyz"
        ...     out = mol.to_xyz(out_file=file)
        ...     with file.open() as f:
        ...         Molecule.from_xyz(f.read()).to_xyz() == out
        True
        """
        geom = "\n".join(map(str, self.atoms))
        if comment is None:
            data = self.model_dump(exclude_none=True, exclude={"atoms"})

            if cell := data.pop("cell", None):
                data["cell"] = cell["lattice_vectors"]
                data["is_periodic"] = cell["is_periodic"]

            comment = " ".join(f"{key}: {value};" for key, value in data.items())

        out = f"{len(self)}\n{comment}\n{geom}"

        if out_file:
            with Path(out_file).open("w") as f:
                f.write(out)

        return out

    @classmethod
    def from_extxyz(cls: type[Self], extxyz: str, charge: int = 0, multiplicity: PositiveInt = 1) -> Self:
        r"""
        Generate a Molecule from a EXTXYZ string. Currently only supporting Lattice and Properties fields.

        >>> Molecule.from_extxyz('''
        ... 2
        ... Lattice="6.0 0.0 0.0 6.0 0.0 0.0 6.0 0.0 0.0"Properties=species:S:1:pos:R:3
        ... H 0 0 0
        ... H 0 0 1
        ... ''').cell.lattice_vectors
        ((6.0, 0.0, 0.0), (6.0, 0.0, 0.0), (6.0, 0.0, 0.0))
        """

        return cls.from_extxyz_lines(extxyz.strip().splitlines(), charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_extxyz_lines(
        cls: type[Self],
        lines: Iterable[str],
        charge: int | None = None,
        multiplicity: PositiveInt | None = None,
        cell: PeriodicCell | None = None,
    ) -> Self:
        """
        Parses an EXTXYZ file, extracting atom positions, forces (if present), and metadata.

        Supports:
        - Lattice vectors (cell information)
        - Properties field (species, positions, forces, etc.)
        - Other metadata like charge, multiplicity, energy, etc.

        :param lines: Iterable of lines from an EXTXYZ file
        :param charge: total charge of the molecule (default: 0 if not found)
        :param multiplicity: spin multiplicity of the molecule (default: 1 if not found)
        :param cell: PeriodicCell containing lattice vectors
        :return: Molecule
        :raises MoleculeReadError: if the file is not in the correct format
        """
        if not isinstance(lines, Sequence):
            lines = list(lines)

        # Ensure first line contains number of atoms
        if len(lines[0].split()) == 1:
            natoms = lines[0].strip()
            if not natoms.isdigit() or (int(natoms) != len(lines) - 2):
                raise MoleculeReadError(f"First line should be number of atoms, got: {lines[0]} != {len(lines) - 2}")
            data_line, *lines = lines[1:]
        else:
            raise MoleculeReadError(f"First line should be an integer denoting atom count. Got {lines[0].split()}")

        metadata = parse_extxyz_comment_line(data_line)

        T = TypeVar("T")

        def metadata_optional_get(key: str, value: T | None, default: T) -> T:
            """Set key to default if not found in metadata"""
            if value is None:
                return metadata.get(key, default)  # type: ignore [return-value]

            return value

        charge = metadata_optional_get("total_charge", charge, 0)
        multiplicity = metadata_optional_get("multiplicity", multiplicity, 1)
        cell = cell or metadata.get("cell")
        energy = metadata.get("energy", None)

        force_idx = None
        if properties := metadata.get("properties", "").split(":"):
            if properties[0].lower() != "species":
                raise MoleculeReadError(f"Invalid or missing 'Properties' field in EXTXYZ, got: {properties}")

            # Identify column indices for position and force data
            pos_idx = None
            current_idx = 0  # Start after 'species:S'

            while current_idx < len(properties):
                if properties[current_idx].lower() == "pos" and properties[current_idx + 1].lower() == "r" and properties[current_idx + 2] == "3":
                    pos_idx = current_idx
                elif properties[current_idx].lower() == "forces" and properties[current_idx + 1].lower() == "r" and properties[current_idx + 2] == "3":
                    force_idx = current_idx
                current_idx += 3

            if pos_idx is None:
                raise MoleculeReadError("No position data ('pos:R:3') found in Properties field.")

        def parse_line_atoms(line: str) -> Atom:
            symbol, sx, sy, sz, *_ = line.split()
            atomic_number = SYMBOL_ELEMENT[symbol.title()]
            x, y, z = map(float, (sx, sy, sz))

            return Atom(atomic_number=atomic_number, position=(x, y, z))

        def parse_line_with_grad(line: str) -> tuple[Atom, Vector3D]:
            symbol, sx, sy, sz, sgx, sgy, sgz, *_ = line.split()
            atomic_number = SYMBOL_ELEMENT[symbol.title()]
            x, y, z = map(float, (sx, sy, sz))
            gx, gy, gz = map(float, (sgx, sgy, sgz))

            return (
                Atom(atomic_number=atomic_number, position=(x, y, z)),
                (-gx, -gy, -gz),
            )

        atoms: list[Atom]
        gradients: list[Vector3D] | None
        if force_idx is not None:
            atoms, gradients = zip(*map(parse_line_with_grad, lines), strict=True)  # type: ignore [assignment]
        else:
            atoms = [parse_line_atoms(line) for line in lines]
            gradients = None

        return cls(atoms=atoms, cell=cell, charge=charge, multiplicity=multiplicity, energy=energy, gradient=gradients)

    @classmethod
    def from_rdkit(cls: type[Self], rdkm: RdkitMol, cid: int = 0, multiplicity: int = 1) -> Self:
        from rdkit import Chem  # noqa: PLC0415

        if len(rdkm.GetConformers()) == 0:
            rdkm = _embed_rdkit_mol(rdkm)

        atomic_numbers = [atom.GetAtomicNum() for atom in rdkm.GetAtoms()]  # type: ignore [no-untyped-call, unused-ignore]
        atoms = [
            Atom(atomic_number=atom, position=xyz)  # keep open
            for atom, xyz in zip(atomic_numbers, rdkm.GetConformers()[cid].GetPositions(), strict=True)
        ]

        charge = Chem.GetFormalCharge(rdkm)

        return cls(atoms=atoms, charge=charge, multiplicity=multiplicity)

    @classmethod
    def from_smiles(cls: type[Self], smiles: str) -> Self:
        from rdkit import Chem  # noqa: PLC0415

        rdkm = Chem.MolFromSmiles(smiles)
        assert rdkm is not None
        return cls.from_rdkit(rdkm)


def _embed_rdkit_mol(rdkm: RdkitMol) -> RdkitMol:
    from rdkit.Chem import AllChem  # noqa: PLC0415

    try:
        AllChem.SanitizeMol(rdkm)  # type: ignore [attr-defined, unused-ignore]
    except Exception as e:
        raise ValueError("Molecule could not be generated -- invalid chemistry!\n") from e

    rdkm = AllChem.AddHs(rdkm)  # type: ignore [attr-defined, unused-ignore]
    try:
        status1 = AllChem.EmbedMolecule(rdkm, maxAttempts=200)  # type: ignore [attr-defined, unused-ignore]
        assert status1 >= 0
    except Exception as e:
        status1 = AllChem.EmbedMolecule(rdkm, maxAttempts=200, useRandomCoords=True)  # type: ignore [attr-defined, unused-ignore]
        if status1 < 0:
            raise ValueError(f"Cannot embed molecule! Error: {e}") from e

    AllChem.MMFFOptimizeMolecule(rdkm, maxIters=200)  # type: ignore [attr-defined, call-arg, unused-ignore]

    return rdkm


class EXTXYZMetadata(TypedDict, total=False):
    """Metadata from EXTXYZ file comment line."""

    properties: Any
    total_charge: int
    multiplicity: int
    energy: float
    cell: PeriodicCell


def parse_extxyz_comment_line(line: str) -> EXTXYZMetadata:
    """
    Parse the comment line of an EXTXYZ file, extracting lattice, properties, and metadata.

    Supports:
    - Lattice vectors (cell information)
    - Properties field (species, positions, forces, etc.)
    - Other metadata fields like charge, multiplicity, energy, etc.

    :param line: comment line from an EXTXYZ file
    :return: parsed properties

    >>> parse_extxyz_comment_line('Lattice="6.0 0.0 0.0 6.0 0.0 0.0 6.0 0.0 0.0"Properties=species:S:1:pos:R:3')
    {'cell': PeriodicCell(lattice_vectors=((6.0, 0.0, 0.0), (6.0, 0.0, 0.0), (6.0, 0.0, 0.0)), is_periodic=(True, True, True)), 'properties': 'species:S:1:pos:R:3'}
    """  # noqa: E501

    # Regular expression to match key="value", key='value', or key=value
    pattern = r"(\S+?=(?:\".*?\"|\'.*?\'|\S+))"
    pairs = re.findall(pattern, line)

    prop_dict: EXTXYZMetadata = {}
    for pair in pairs:
        key, value = pair.split("=", 1)
        key = key.lower().strip()
        value = value.strip("'\"")

        if key == "lattice":
            lattice_values = value.split()
            if len(lattice_values) != 9:
                raise MoleculeReadError(f"Lattice should have 9 entries, got {len(lattice_values)}")

            try:
                cell = tuple(tuple(map(float, lattice_values[i : i + 3])) for i in range(0, 9, 3))
            except ValueError as e:
                raise MoleculeReadError(f"Lattice should be floats, got {lattice_values}") from e

            prop_dict["cell"] = PeriodicCell(lattice_vectors=cell)

        elif key == "properties":
            prop_dict["properties"] = value

        elif key == "total_charge":
            prop_dict["total_charge"] = int(value)
        elif key == "multiplicity":
            prop_dict["multiplicity"] = int(value)
        elif key == "energy":
            prop_dict["energy"] = float(value)
        else:
            prop_dict[key] = value  # type: ignore [literal-required]

    return prop_dict


def angle(p0: Sequence[float], p1: Sequence[float], p2: Sequence[float], degrees: bool = True) -> float:
    """
    Angle between three points.

    :param p0: position of first point
    :param p1: position of vertex point
    :param p2: position of third point
    :param degrees: whether to return in degrees
    :return: angle in radians or degrees
    """
    a0, a1, a2 = map(np.asarray, (p0, p1, p2))
    u = a1 - a0
    v = a1 - a2

    nu = np.linalg.norm(u)
    nv = np.linalg.norm(v)
    cos_theta = np.dot(u, v) / (nu * nv)
    cos_theta = np.clip(cos_theta, -1.0, 1.0)

    ang = np.arccos(cos_theta)
    if degrees:
        return float(np.degrees(ang))
    return float(ang)


def dihedral(p0: Sequence[float], p1: Sequence[float], p2: Sequence[float], p3: Sequence[float], degrees: bool = True, positive_domain: bool = True) -> float:
    """
    Dihedral angle between four points.

    :param p0, p1, p2, p3: points
    :param degrees: whether to return in degrees
    :param positive_domain: (0, 360] if True else (-180, 180]
    :return: angle in degrees or radians (or nan if collinearities detected)

    >>> a = [0, 0, 0]
    >>> b = [0, 0, 1]
    >>> c = [0, 1, 1]
    >>> d1 = [0, 1, 2]
    >>> d2 = [0.5, 1, 1.5]
    >>> dihedral(a, b, c, d1)
    180.0
    >>> dihedral(a, b, c, d2, positive_domain=False)
    -135.0
    """
    a0, a1, a2, a3 = map(np.asarray, (p0, p1, p2, p3))
    b0 = a1 - a0
    b1 = a2 - a1
    b2 = a3 - a2

    b1 = b1 / np.linalg.norm(b1)

    v = b1 * np.dot(b0, b1) - b0
    w = b2 - b1 * np.dot(b2, b1)

    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    ang = np.arctan2(y, x)

    if positive_domain and ang < 0:
        ang += 2 * np.pi

    if degrees:
        return float(np.degrees(ang))
    return float(ang)
